"""Basic pdpipe PdPipelineStages."""

from typing import Optional, List, Dict, Union, Callable
from collections import deque

import pandas
from strct.dicts import reverse_dict_partial

from pdpipe.core import PdPipelineStage, ColumnsBasedPipelineStage
# from pdpipe.util import out_of_place_col_insert
from pdpipe.shared import (
    _interpret_columns_param,
    _list_str,
)

import pdpipe.cond as cond
from pdpipe.types import ColumnsParamType
from pdpipe.exceptions import FailedConditionError
from pdpipe.cq import ColumnQualifier


class ColDrop(ColumnsBasedPipelineStage):
    """A pipeline stage that drops columns by name.

    Parameters
    ----------
    columns : single label, list-like or callable
        The label, or an iterable of labels, of columns to drop. Alternatively,
        this parameter can be assigned a callable returning an iterable of
        labels from an input pandas.DataFrame (see `pdpipe.cq`).
    errors : {‘ignore’, ‘raise’}, default ‘raise’
        If ‘ignore’, suppress error and existing labels are dropped.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[8,'a'],[5,'b']], [1,2], ['num', 'char'])
    >>> pdp.ColDrop('num').apply(df)
      char
    1    a
    2    b
    """

    def __init__(
        self,
        columns: ColumnsParamType,
        errors: Optional[str] = None,
        **kwargs: object,
    ) -> None:
        self._errors = errors
        self._post_cond = cond.HasNoColumn(columns)
        super_kwargs = {
            'columns': columns,
            'desc_temp': 'Drop columns {}',
        }
        super_kwargs.update(**kwargs)
        super_kwargs['none_columns'] = 'error'
        super().__init__(**super_kwargs)

    def _prec(self, X: pandas.DataFrame) -> bool:
        if self._errors != 'ignore':
            return super()._prec(X)
        return True

    def _post(self, X: pandas.DataFrame) -> bool:
        return self._post_cond(X)

    def _transformation(
        self, X: pandas.DataFrame, verbose: bool, fit: bool,
    ) -> pandas.DataFrame:
        return X.drop(
            self._get_columns(X, fit=fit), axis=1, errors=self._errors)


class ValDrop(ColumnsBasedPipelineStage):
    """A pipeline stage that drops rows by value.

    Parameters
    ----------
    values : list-like
        A list of the values to drop.
    columns : single label, list-like or callable, default None
        The label, or an iterable of labels, of columns to check for the given
        values. Alternatively, this parameter can be assigned a callable
        returning an iterable of labels from an input pandas.DataFrame. See
        `pdpipe.cq`. If set to None, all columns are checked.
    exclude_columns : label, iterable or callable, optional
        The label, or an iterable of labels, of columns to exclude, given the
        `columns` parameter. Alternatively, this parameter can be assigned a
        callable returning a labels iterable from an input pandas.DataFrame.
        See `pdpipe.cq`. Optional. By default no columns are excluded.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[1,4],[4,5],[18,11]], [1,2,3], ['a','b'])
    >>> pdp.ValDrop([4], 'a').apply(df)
        a   b
    1   1   4
    3  18  11
    >>> pdp.ValDrop([4]).apply(df)
        a   b
    3  18  11
    """

    def __init__(
        self,
        values: List[object],
        columns: ColumnsParamType = None,
        **kwargs: object,
    ) -> None:
        self._values = values
        self._values_str = _list_str(self._values)
        super_kwargs = {
            'columns': columns,
            'desc_temp': f'Drop values {self._values_str} in columns {{}}',
        }
        super_kwargs.update(**kwargs)
        super_kwargs['none_columns'] = 'all'
        super().__init__(**super_kwargs)

    def _transformation(
        self, X: pandas.DataFrame, verbose: bool, fit: bool,
    ) -> pandas.DataFrame:
        inter_X = X
        before_count = len(inter_X)
        columns_to_check = self._get_columns(X, fit=fit)
        for col in columns_to_check:
            inter_X = inter_X[~inter_X[col].isin(self._values)]
        if verbose:
            print(f"{before_count - len(inter_X)} rows dropped.")
        return inter_X


class ValKeep(ColumnsBasedPipelineStage):
    """A pipeline stage that keeps rows by value.

    Parameters
    ----------
    values : list-like
        A list of the values to keep.
    columns : single label, list-like or callable, default None
        The label, or an iterable of labels, of columns to check for the given
        values. Alternatively, this parameter can be assigned a callable
        returning an iterable of labels from an input pandas.DataFrame. See
        `pdpipe.cq`. If set to None, all columns are checked.
    exclude_columns : single label, iterable or callable, optional
        The label, or an iterable of labels, of columns to exclude, given the
        `columns` parameter. Alternatively, this parameter can be assigned a
        callable returning a labels iterable from an input pandas.DataFrame.
        See `pdpipe.cq`. Optional. By default no columns are excluded.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[1,4],[4,5],[5,11]], [1,2,3], ['a','b'])
    >>> pdp.ValKeep([4, 5], 'a').apply(df)
       a   b
    2  4   5
    3  5  11
    >>> pdp.ValKeep([4, 5]).apply(df)
       a  b
    2  4  5
    """

    def __init__(self, values, columns=None, **kwargs):
        self._values = values
        self._values_str = _list_str(self._values)
        super_kwargs = {
            'columns': columns,
            'desc_temp': f'Keep values {self._values_str} in columns {{}}',
        }
        super_kwargs.update(**kwargs)
        super_kwargs['none_columns'] = 'all'
        super().__init__(**super_kwargs)

    def _transformation(self, X, verbose, fit):
        inter_X = X
        before_count = len(inter_X)
        columns_to_check = self._get_columns(X, fit=fit)
        for col in columns_to_check:
            inter_X = inter_X[inter_X[col].isin(self._values)]
        if verbose:
            print(f"{before_count - len(inter_X)} rows dropped.")
        return inter_X


class ColRename(PdPipelineStage):
    """A pipeline stage that renames a column or columns.

    Parameters
    ----------
    rename_mapper : dict-like or callable
        Maps old column names to new ones.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[8,'a'],[5,'b']], [1,2], ['num', 'char'])
    >>> pdp.ColRename({'num': 'len', 'char': 'initial'}).apply(df)
       len initial
    1    8       a
    2    5       b

    >>> def renamer(lbl: str):
    ...    if lbl.startswith('n'):
    ...       return 'foo'
    ...    return lbl
    >>> pdp.ColRename(renamer).apply(df)
       foo char
    1    8    a
    2    5    b
    """

    _DEF_COLDRENAME_EXC_MSG = ("ColRename stage failed because not all columns"
                               " {} were found in input dataframe.")

    def __init__(self, rename_mapper: Union[Dict, Callable], **kwargs):
        self._rename_mapper = rename_mapper
        try:
            columns_str = _list_str(list(rename_mapper.keys()))
            mapper_repr = str(rename_mapper)
            keys_set = set(self._rename_mapper.keys())
            required_labels = list(keys_set)
            _tprec = cond.HasAllColumns(required_labels)
        except AttributeError:  # rename mapper is a callable
            mapper_repr = rename_mapper.__name__
            doc = rename_mapper.__doc__
            if doc is None:
                columns_str = f"by func {rename_mapper.__name__}"
            else:
                columns_str = (
                    f"by func {rename_mapper.__name__} with "
                    f"doc: {rename_mapper.__doc__}"
                )
            _tprec = cond.AlwaysTrue()
        try:
            suffix = 's' if len(rename_mapper) > 1 else ''
        except TypeError:
            suffix = 's'
        self._tprec = _tprec
        super_kwargs = {
            'exmsg': ColRename._DEF_COLDRENAME_EXC_MSG.format(columns_str),
            'desc': f"Rename column{suffix} with {mapper_repr}",
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, X):
        return self._tprec(X)

    def _transform(self, X, verbose):
        return X.rename(columns=self._rename_mapper)


class DropNa(PdPipelineStage):
    """A pipeline stage that drops null values.

    Supports all parameter supported by pandas.dropna function.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[1,4],[4,None],[1,11]], [1,2,3], ['a','b'])
    >>> pdp.DropNa().apply(df)
       a     b
    1  1   4.0
    3  1  11.0
    """

    _DEF_DROPNA_EXC_MSG = "DropNa stage failed."
    _DROPNA_KWARGS = ['axis', 'how', 'thresh', 'subset', 'inplace']

    def __init__(self, **kwargs):
        common = set(kwargs.keys()).intersection(DropNa._DROPNA_KWARGS)
        self.dropna_kwargs = {key: kwargs.pop(key) for key in common}
        super_kwargs = {
            'exmsg': DropNa._DEF_DROPNA_EXC_MSG,
            'desc': "Drops null values."
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, X):
        return True

    def _transform(self, X, verbose):
        before_count = len(X)
        ncols_before = len(X.columns)
        inter_X = X.dropna(**self.dropna_kwargs)
        if verbose:
            print(
                f"{before_count - len(inter_X)} rows, "
                f"{ncols_before - len(inter_X.columns)} columns dropeed"
            )
        return inter_X


class SetIndex(PdPipelineStage):
    """A pipeline stage that set existing columns as index.

    Supports all parameter supported by pandas.set_index function except for
    `inplace`.

    Example
    -------
    >> import pandas as pd; import pdpipe as pdp;
    >> df = pd.DataFrame([[1,4],[3, 11]], [1,2], ['a','b'])
    >> pdp.SetIndex('a').apply(df)
        b
    a
    1   4
    3  11
    """

    _DEF_SETIDX_EXC_MSG = "SetIndex stage failed."
    _DEF_SETIDX_APP_MSG = "Setting indexes..."
    _SETINDEX_KWARGS = ['drop', 'append', 'verify_integrity']

    def __init__(self, keys, **kwargs):
        common = set(kwargs.keys()).intersection(SetIndex._SETINDEX_KWARGS)
        self.setindex_kwargs = {key: kwargs.pop(key) for key in common}
        self.keys = keys
        if hasattr(keys, '__iter__') and not isinstance(keys, str):
            _tprec = cond.HasAllColumns(list(keys))
        else:
            _tprec = cond.HasAllColumns([keys])
        self._tprec = _tprec
        super_kwargs = {
            'exmsg': SetIndex._DEF_SETIDX_EXC_MSG,
            'desc': "Set indexes."
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, X):
        return self._tprec(X)

    def _transform(self, X, verbose):
        return X.set_index(keys=self.keys, **self.setindex_kwargs)


class FreqDrop(PdPipelineStage):
    """A pipeline stage that drops rows by value frequency.

    Parameters
    ----------
    threshold : int
        The minimum frequency required for a value to be kept.
    column : str
        The name of the colum to check for the given value frequency.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[1,4],[4,5],[1,11]], [1,2,3], ['a','b'])
    >>> pdp.FreqDrop(2, 'a').apply(df)
       a   b
    1  1   4
    3  1  11
    """

    _DEF_FREQDROP_EXC_MSG = ("FreqDrop stage failed because column {} was not"
                             " found in input dataframe.")
    _DEF_FREQDROP_DESC = "Drop values with frequency < {} in column {}."

    def __init__(self, threshold: int, column: str, **kwargs):
        self._threshold = threshold
        self._column = column
        super_kwargs = {
            'exmsg': FreqDrop._DEF_FREQDROP_EXC_MSG.format(self._column),
            'desc': FreqDrop._DEF_FREQDROP_DESC.format(
                self._threshold, self._column)
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, X):
        return self._column in X.columns

    def _transform(self, X, verbose):
        inter_X = X
        before_count = len(inter_X)
        valcount = X[self._column].value_counts()
        to_drop = valcount[valcount < self._threshold].index
        inter_X = inter_X[~inter_X[self._column].isin(to_drop)]
        if verbose:
            print(f"{before_count - len(inter_X)} rows dropped.")
        return inter_X


class ColReorder(PdPipelineStage):
    """A pipeline stage that reorders columns.

    Parameters
    ----------
    positions : dict
        A mapping of column names to their desired positions after reordering.
        Columns not included in the mapping will maintain their relative
        positions over the non-mapped colums.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[8,4,3,7]], columns=['a', 'b', 'c', 'd'])
    >>> pdp.ColReorder({'b': 0, 'c': 3}).apply(df)
       b  a  d  c
    0  4  8  7  3
    """

    _DEF_ORD_EXC_MSG = ("ColReorder stage failed because not all columns in {}"
                        " were found in input dataframe.")

    def __init__(self, positions, **kwargs):
        self._col_to_pos = positions
        self._pos_to_col = reverse_dict_partial(positions)
        super_kwargs = {
            'exmsg': ColReorder._DEF_ORD_EXC_MSG.format(self._col_to_pos),
            'desc': f"Reorder columns by {self._col_to_pos}",
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, X):
        return set(self._col_to_pos.keys()).issubset(X.columns)

    def _transform(self, X, verbose):
        cols = X.columns
        map_cols = list(self._col_to_pos.keys())
        non_map_cols = deque(x for x in cols if x not in map_cols)
        new_columns = []
        try:
            for pos in range(len(cols)):
                if pos in self._pos_to_col:
                    new_columns.append(self._pos_to_col[pos])
                else:
                    new_columns.append(non_map_cols.popleft())
            return X[new_columns]
        except (IndexError):
            raise ValueError(f"Bad positions mapping given: {new_columns}")


class RowDrop(ColumnsBasedPipelineStage):
    """A pipeline stage that drops rows by callable conditions.

    Parameters
    ----------
    conditions : list-like or dict
        The list of conditions that make a row eligible to be dropped. Each
        condition must be a callable that take a cell value and return a bool
        value. If a list of callables is given, the conditions are checked for
        each column value of each row. If a dict mapping column labels to
        callables is given, then each condition is only checked for the column
        values of the designated column.
    reduce : 'any', 'all' or 'xor', default 'any'
        Determines how row conditions are reduced. If set to 'all', a row must
        satisfy all given conditions to be dropped. If set to 'any', rows
        satisfying at least one of the conditions are dropped. If set to 'xor',
        rows satisfying exactly one of the conditions will be dropped. Set to
        'any' by default.
    columns : single label, iterable or callable, optional
        The label, or an iterable of labels, of columns. Alternatively,
        this parameter can be assigned a callable returning an iterable of
        labels from an input pandas.DataFrame. See `pdpipe.cq`. If given,
        input conditions will be applied to the sub-dataframe made up of
        these columns to determine which rows to drop. Ignored if `conditions`
        is provided with a dict object. If `conditions` is a list and this
        parameter is not provided, all columns are checked (unless
        `exclude_columns` is additionally provided)
    exclude_columns : single label, iterable or callable, optional
        The label, or an iterable of labels, of columns to exclude, given the
        `columns` parameter. Alternatively, this parameter can be assigned a
        callable returning a labels iterable from an input pandas.DataFrame.
        See `pdpipe.cq`. Optional. By default no columns are excluded.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[1,4],[4,5],[5,11]], [1,2,3], ['a','b'])
    >>> pdp.RowDrop([lambda x: x < 2]).apply(df)
       a   b
    2  4   5
    3  5  11
    >>> pdp.RowDrop({'a': lambda x: x == 4}).apply(df)
       a   b
    1  1   4
    3  5  11
    """

    _REDUCERS = {
        'all': all,
        'any': any,
        'xor': lambda x: sum(x) == 1
    }

    class _DictRowCond(object):
        """Filter rows by a dict of conditions."""

        def __init__(self, conditions, reducer):
            self.conditions = conditions
            self.reducer = reducer

        def __call__(self, row):
            res = [cond(row[lbl]) for lbl, cond in self.conditions.items()]
            return self.reducer(res)

    class _ListRowCond(object):
        """Filter rows by a list of conditions."""

        def __init__(self, conditions, reducer):
            self.conditions = conditions
            self.reducer = reducer

        def __call__(self, row):
            res = [self.reducer(row.apply(cond)) for cond in self.conditions]
            return self.reducer(res)

    def _row_condition_builder(self, conditions, reduce):
        reducer = RowDrop._REDUCERS[reduce]
        if self._cond_is_dict:
            row_cond = RowDrop._DictRowCond(
                conditions=conditions, reducer=reducer)
        else:
            row_cond = RowDrop._ListRowCond(
                conditions=conditions, reducer=reducer)
        return row_cond

    def __init__(self, conditions, reduce=None, columns=None, **kwargs):
        self._conditions = conditions
        if reduce is None:
            reduce = 'any'
        self._reduce = reduce
        if reduce not in RowDrop._REDUCERS.keys():
            raise ValueError((
                "{} is an unsupported argument for the 'reduce' parameter of "
                "the RowDrop constructor!").format(reduce))
        self._cond_is_dict = isinstance(conditions, dict)
        if self._cond_is_dict:
            valid = all([callable(cond) for cond in conditions.values()])
            if not valid:
                raise ValueError(
                    "Condition dicts given to RowDrop must map to callables!")
            columns = list(conditions.keys())
        else:
            valid = all([callable(cond) for cond in conditions])
            if not valid:
                raise ValueError(
                    "RowDrop condition lists can contain only callables!")
        self._row_cond = self._row_condition_builder(conditions, reduce)
        super_kwargs = {
            'columns': columns,
            'desc_temp': 'Drop rows in columns {} by conditions',
        }
        super_kwargs.update(**kwargs)
        super_kwargs['none_columns'] = 'all'
        super().__init__(**super_kwargs)

    def _transformation(self, X, verbose, fit):
        before_count = len(X)
        columns = self._get_columns(X, fit=fit)
        subX = X[columns]
        drop_index = ~subX.apply(self._row_cond, axis=1)
        inter_X = X[drop_index]
        if verbose:
            print(f"{before_count - len(inter_X)} rows dropped.")
        return inter_X


class Schematize(PdPipelineStage):
    """Enforces a column schema on input dataframes.

    Parameters
    ----------
    columns: sequence of label, optional
        The dataframe schema to enforce on input dataframes. If set to None,
        the schema is learned in fit time and applied in subsequent transforms.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[2, 4, 8],[3, 6, 9]], [1, 2], ['a', 'b', 'c'])
    >>> pdp.Schematize(['a', 'c']).apply(df)
       a  c
    1  2  8
    2  3  9
    >>> pdp.Schematize(['c', 'b']).apply(df)
       c  b
    1  8  4
    2  9  6
    """

    def __init__(
        self,
        columns: Optional[List[object]],
        **kwargs: object,
    ) -> None:
        if columns is None:
            self._adaptive = True
            self._columns = None
            self._columns_str = '<Learnable Schema>'
            exmsg = "Learnable schematize failed in precondition unexpectedly!"
        else:
            self._adaptive = False
            self._columns = _interpret_columns_param(columns)
            self._columns_str = _list_str(self._columns)
            exmsg = (
                f"Not all required columns {self._columns_str} "
                f"found in input dataframe!"
            )
        desc = (
            f"Transform input dataframes to the following schema: "
            f"{self._columns_str}"
        )
        super_kwargs = {
            'exmsg': exmsg,
            'desc': desc,
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, X: pandas.DataFrame) -> bool:
        if self._adaptive and not self.is_fitted:
            return True
        return set(self._columns).issubset(X.columns)

    def _transform(
            self, X: pandas.DataFrame, verbose=None) -> pandas.DataFrame:
        return X[self._columns]

    def _fit_transform(
            self, X: pandas.DataFrame, verbose=None) -> pandas.DataFrame:
        if self._adaptive:
            self._columns = X.columns
            self.is_fitted = True
            return X
        return X[self._columns]


class DropDuplicates(ColumnsBasedPipelineStage):
    """Drop duplicates in the given columns.

    Parameters
    ----------
    columns: column label or sequence of labels, optional
        The labels of the columns to consider for duplication drop. If not
        populated, duplicates are dropped from all columns.
    exclude_columns : object, iterable or callable, optional
        The label, or an iterable of labels, of columns to exclude, given the
        `columns` parameter. Alternatively, this parameter can be assigned a
        callable returning a labels iterable from an input pandas.DataFrame.
        See `pdpipe.cq`. Optional. By default no columns are excluded.

    Examples
    --------
        >>> import pandas as pd; import pdpipe as pdp;
        >>> df = pd.DataFrame([[8, 1],[8, 2], [9, 2]], [1,2,3], ['a', 'b'])
        >>> pdp.DropDuplicates('a').apply(df)
           a  b
        1  8  1
        3  9  2
    """

    def __init__(self, columns=None, **kwargs):
        super_kwargs = {
            'columns': columns,
            'desc_temp': 'Drop duplicates in columns {}',
        }
        super_kwargs.update(**kwargs)
        super_kwargs['none_columns'] = 'all'
        super().__init__(**super_kwargs)

    def _transformation(self, X, verbose, fit):
        columns = self._get_columns(X, fit=fit)
        inter_X = X.drop_duplicates(subset=columns)
        if verbose:
            print(f"{len(X) - len(inter_X)} rows dropped.")
        return inter_X


class ColumnDtypeEnforcer(PdPipelineStage):
    """A pipeline stage enforcing column dtypes.

    Parameters
    ----------
    column_to_dtype: dict of labels / ColumnQualifiers to dtypes
        Use {col: dtype, …}, where col is a column label and dtype is a
        numpy.dtype or Python type to cast one or more of the DataFrame’s
        columns to column-specific types. Alternatively, you can provide
        `ColumnQualifier` objects as keys. If at least one such key is present,
        the lbl-to-dtype dict is dynamically inferred each time the pipeline
        stage is applied (note that `ColumnQualifier` objects are fittable by
        default, so to have column labels re-inferred after the first stage
        application you'll have to set `fittable=False` for the
        `ColumnQualifier` you use, see `pdpipe.cq`).
    errors: {‘raise’, ‘ignore’}, default ‘raise’
        Control raising of exceptions on invalid data for provided dtype.
        - raise : allow exceptions to be raised
        - ignore : suppress exceptions. On error return original object.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[8,'a'],[5,'b']], [1,2], ['num', 'initial'])
    >>> pdp.ColumnDtypeEnforcer({'num': float}).apply(df)
       num initial
    1  8.0       a
    2  5.0       b

    >>> pdp.ColumnDtypeEnforcer({pdp.cq.StartWith('n'): float}).apply(df)
       num initial
    1  8.0       a
    2  5.0       b
    """

    _DEF_COL_DTYPE_ENF_EXC_MSG = (
        "ColumnDtypeEnforcer stage failed because not all columns"
        " {} were found in input dataframe.")

    def __init__(
        self,
        column_to_dtype: Dict,
        errors: Optional[str] = 'raise',
        **kwargs: object,
    ) -> None:
        # if none of the keys in column_to_dtype is a ColumnQualifier
        if not any(isinstance(
                x, ColumnQualifier) for x in column_to_dtype.keys()):
            # its a static map; use it as is
            self._column_to_dtype = column_to_dtype
            keys_set = set(column_to_dtype.keys())
            _tprec = cond.HasAllColumns(list(keys_set))
        else:
            # else, it's at least partly dynamic, and will have to infer it
            # on run time
            self._dynamic_column_to_dtype = column_to_dtype
            _tprec = cond.AlwaysTrue()
        self._tprec = _tprec
        self._errors = errors
        columns_str = _list_str(list(column_to_dtype.keys()))
        suffix = 's' if len(column_to_dtype) > 1 else ''
        super_kwargs = {
            'exmsg': ColumnDtypeEnforcer._DEF_COL_DTYPE_ENF_EXC_MSG.format(
                columns_str),
            'desc': f"Enforce column{suffix} dtype with {column_to_dtype}",
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _col_to_dtype_from_X(self, X: pandas.DataFrame) -> Dict:
        try:
            return self._column_to_dtype
        except AttributeError:
            column_to_dtype = {}
            for k, dtype in self._dynamic_column_to_dtype.items():
                try:
                    column_to_dtype.update({
                        lbl: dtype
                        for lbl in k(X)
                    })
                except TypeError:  # k is not a callable
                    column_to_dtype[k] = dtype
            return column_to_dtype

    def _prec(self, X: pandas.DataFrame) -> bool:
        return self._tprec(X)

    def _transform(
        self,
        X: pandas.DataFrame,
        verbose: bool,
    ) -> pandas.DataFrame:
        lbl_to_dtype = self._col_to_dtype_from_X(X)
        return X.astype(
            dtype=lbl_to_dtype,
            copy=True,
            errors=self._errors,
        )


class ConditionValidator(PdPipelineStage):
    """A pipeline stage that validates boolean conditions on dataframes.

    The stage does not change the input dataframe in any way.

    The constructor expects either a single callable or a list-like of callable
    objects, and checks that all these callable return True - meaning all
    defined conditions hold - for input dataframes.

    Naturally, pdpipe `Condition` objects from the `pdpipe.cond` module can be used.

    Parameters
    ----------
    conditions : callable or list-like of callable
        The conditions to check for input dataframes. Naturally, pdpipe
        `Condition` objects from the `pdpipe.cond` module can be used.
    reducer : callable, optional
        The callable that reduces the list of boolean result to a single
        result. By default the built-in `all` function is used, so all
        conditions must hold for this pipeline stage to validate an input
        dataframe. The built-in `any` function may be used to validate at least
        one condition holds, and of course custom reducing functions can be
        used.
    errors : str, default 'raise'
        If set to 'raise', the default, then if the result boolean result is
        False a FailedConditionError is raised on stage application. If set to
        'ignore', then conditions are checked, the results are printed if the
        application was called with `verbose=True`, and pipeline application
        continues. Any other value is interpreted as 'raise'.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[1,4],[4,None],[1,11]], [1,2,3], ['a','b'])
    >>> pdp.ConditionValidator(lambda df: len(df.columns) == 5).apply(df)
    Traceback (most recent call last):
       ...
    pdpipe.exceptions.FailedConditionError: ConditionValidator stage failed; some conditions did not hold for the input dataframe!

    >>> pdp.ConditionValidator(pdp.cond.HasNoMissingValues()).apply(df)
    Traceback (most recent call last):
       ...
    pdpipe.exceptions.FailedConditionError: ConditionValidator stage failed; some conditions did not hold for the input dataframe!
    """  # noqa: E501

    def __init__(
        self,
        conditions: Union[Callable, List[Callable]],
        reducer: Optional[Callable] = all,
        errors: Optional[str] = 'raise',
        **kwargs: object,
    ):
        if callable(conditions):
            conditions = [conditions]
        self._conditions = conditions
        self._reducer = reducer
        self._errors = 'raise'
        self._raise = True
        if errors == 'ignore':
            self._errors = 'ignore'
            self._raise = False
        super_kwargs = {
            'desc': "Validates conditions"
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, X):
        return True

    def _transform(self, X, verbose):
        results = []
        for cond_obj in self._conditions:
            try:
                res = cond_obj(X)
            except Exception as e:
                raise ValueError(
                    f"Supplied condition raised a {e} exception when applied "
                    "to input dataframe!"
                ) from e
            if verbose and not res:
                cond_repr = cond_obj.__doc__
                if cond_repr is None:
                    cond_repr = str(cond_obj)
                print(f"  + Condition failed for input dataframe: {cond_repr}")
            results.append(res)
        reduced_result = self._reducer(results)
        if self._raise and not reduced_result:
            raise FailedConditionError(
                "ConditionValidator stage failed; some conditions did not hold"
                " for the input dataframe!")
        return X


class ApplicationContextEnricher(PdPipelineStage):
    """A pipeline stage that enriches the pipeline's application context.

    Keyword arguments can be either PdPipelineStage constructor arguments, in
    which case they are passed to the stage constructor, or they can be
    mappings to be added to the application context. If a key maps to a
    callable, then the callable is called with the input dataframe as the first
    argument, and the application context is passed as a keyword argument if
    the callable expects it (otherwise only the input dataframe is passed),
    and the result is stored in the application context mapped by the key. If
    the key maps to a non-callable object, the mapping is simply stored in the
    application context.

    Mappings are evaluated in the order they are passed to the constructor.

    For example, `ApplicationContextEnricher(suma=lambda df: df['a'].sum())`
    will add the sum of the 'a' column to the application context keys under
    the key 'suma'. Lated stages can then access the value of 'suma' with
    `self.application_context['suma']`.

    Similarly, `ApplicationContextEnricher(b=5)` will add the {'b': 5} mapping
    to the application context.

    Gradual mappings that use earlier mappings should be given in order, e.g.:
    ```
    ApplicationContextEnricher(
        asum=lambda df: df['a'].sum(),
        bsum=lambda df: df['a'].mean(),
        absumdiff=lambda df, application_context: application_context['asum'] - application_context['bsum'],
    )
    ```
    If all three quantities are required by later stages, and computation time
    for the intermediate quantities is substantial, then this approach can save
    redundant computation.

    Parameters
    ----------
    **kwargs : str to object mapping
        The mappings to be added to the application context.
    """  # noqa: E501

    def __init__(
        self,
        **kwargs: object,
    ):
        init_kwargs, enrichments = self._split_kwargs(kwargs)
        self._enrichments = enrichments
        super_kwargs = {
            'desc': "Enrich application context"
        }
        super_kwargs.update(**init_kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, X):
        return True

    def _transform(self, X, verbose):
        for k, v in self._enrichments.items():
            if callable(v):
                try:
                    self.application_context[k] = v(
                        X, application_context=self.application_context)
                except TypeError:
                    try:
                        self.application_context[k] = v(X)
                    except Exception as e:
                        raise ValueError((
                            f"Supplied enrichment function raised a {e} "
                            "exception when applied to input dataframe!"
                        )) from e
            else:
                self.application_context[k] = v
        return X



"""Text processing pdpipe pipeline stages."""

import re
from typing import Optional

from pdpipe.col_generation import ApplyByCols
from pdpipe.types import ColumnsParamType, ColumnLabelsType


class RegexReplace(ApplyByCols):
    """A pipeline stage replacing regex occurences in a text column.

    Parameters
    ----------
    columns : single label, list-like or callable
        Column labels in the DataFrame which regex replacement be applied to.
        Alternatively, this parameter can be assigned a callable returning an
        iterable of labels from an input pandas.DataFrame. See `pdpipe.cq`.
    pattern : str
        The regex whose occurences will be replaced.
    replace : str
        The replacement string to use. This is equivalent to repl in re.sub.
    flags : int, default 0
        Regex flags that are compatible with Python's `re` module.
    result_columns : label or list-like of labels, default None
        The labels of the new columns resulting from the mapping operation.
        Must be of the same length as columns. If None, behavior depends on the
        drop parameter: If drop is True, the label of the source column is
        used; otherwise, the label of the source column is casted to a string
        and concatenated with the suffix '_reg'.
    drop : bool, default True
        If set to True, source columns are dropped after being transformed.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp; import re;
    >>> data = [[4, "more than 12"], [5, "with 5 more"]]
    >>> df = pd.DataFrame(data, [1,2], ["age","text"])
    >>> clean_num = pdp.RegexReplace('text', r'\\b[0-9]+\\b', "NUM")
    >>> clean_num(df)
       age           text
    1    4  more than NUM
    2    5  with NUM more

    >>> data = [["Mr. John", 18], ["MR. Bob", 25]]
    >>> df = pd.DataFrame(data, [1,2], ["name","age"])
    >>> match_men = r'^mr.*'
    >>> censor_men = pdp.RegexReplace(
    ...    'name', match_men, "x", flags=re.IGNORECASE
    ... )
    >>> censor_men(df)
      name  age
    1    x   18
    2    x   25
    """  # noqa: W605

    class RegexReplacer(object):
        """A pickle-able regex replacement function."""

        def __init__(
            self,
            pattern_str: str,
            replace_text: str,
            flags: Optional[int] = 0,
        ) -> None:
            self.pattern_str = pattern_str
            self.replace_text = replace_text
            self.flags = flags
            self.pattern_obj = re.compile(pattern_str, flags=flags)

        def __call__(self, string: str):
            return self.pattern_obj.sub(self.replace_text, string)

    def __init__(
        self,
        columns: ColumnsParamType,
        pattern: str,
        replace: str,
        flags: Optional[int] = 0,
        result_columns: Optional[ColumnLabelsType] = None,
        drop: Optional[bool] = True,
        func_desc: Optional[str] = None,
        **kwargs,
    ):
        self._pattern_str = pattern
        self._replace = replace
        self._flags = flags
        desc_temp = "Replacing appearances of {} with '{}' in column {{}}"
        desc_temp = desc_temp.format(pattern, replace)
        super_kwargs = {
            'columns': columns,
            'func': RegexReplace.RegexReplacer(
                self._pattern_str, self._replace, self._flags),
            'suffix': '_regex',
            'result_columns': result_columns,
            'drop': drop,
            'desc_temp': desc_temp,
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)


class DropTokensByLength(ApplyByCols):
    """A pipeline stage removing tokens by length in string-token list columns.

    Parameters
    ----------
    columns : single label, list-like or callable
        Names of token list columns on which to apply token filtering.
        Alternatively, this parameter can be assigned a callable returning an
        iterable of labels from an input pandas.DataFrame. See `pdpipe.cq`.
    min_len : int
        The minimum length of tokens to keep. Tokens of shorter length are
        removed from all token lists.
    max_len : int, default None
        The maximum length of tokens to keep. If provided, tokens of longer
        length are removed from all token lists.
    result_columns : str or list-like, default None
        The names of the new columns resulting from the mapping operation.
        Must be of the same length as columns. If None, behavior depends on
        the drop parameter: If drop is True, the name of the source column
        is used; otherwise, the name of the source column is used with the
        suffix '_filtered'.
    drop : bool, default True
        If set to True, source columns are dropped after being transformed.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> data = [[4, ["a", "bad", "nice"]], [5, ["good", "university"]]]
    >>> df = pd.DataFrame(data, [1,2], ["age","text"])
    >>> filter_tokens = pdp.DropTokensByLength('text', 3, 5)
    >>> filter_tokens(df)
       age         text
    1    4  [bad, nice]
    2    5       [good]
    """  # noqa: W605

    class MinLengthTokenFilter(object):

        def __init__(self, min_len):
            self.min_len = min_len

        def __call__(self, token_list):
            return [x for x in token_list if len(x) >= self.min_len]

    class MinMaxLengthTokenFilter(object):

        def __init__(self, min_len, max_len):
            self.min_len = min_len
            self.max_len = max_len

        def __call__(self, token_list):
            return [
                x for x in token_list
                if len(x) >= self.min_len and len(x) <= self.max_len
            ]

    def __init__(
        self, columns, min_len, max_len=None, result_columns=None, drop=True,
        **kwargs
    ):
        self._min_len = min_len
        self._max_len = max_len
        token_filter = DropTokensByLength.MinLengthTokenFilter(min_len)
        cond_str = f" > {min_len}"
        if max_len:
            token_filter = DropTokensByLength.MinMaxLengthTokenFilter(
                min_len=min_len, max_len=max_len)
            cond_str += f" < {max_len}"
        desc_temp = "Filtering out tokens of length{} in columns {{}}"
        desc_temp = desc_temp.format(cond_str)
        super_kwargs = {
            'columns': columns,
            'func': token_filter,
            'result_columns': result_columns,
            'drop': drop,
            'suffix': "_filtered",
            'desc_temp': desc_temp,
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)


class DropTokensByList(ApplyByCols):
    """A pipeline stage removing specific tokens in string-token list columns.

    Parameters
    ----------
    columns : single label, list-like or callable
        Names of token list columns on which to apply token filtering.
        Alternatively, this parameter can be assigned a callable returning an
        iterable of labels from an input pandas.DataFrame. See `pdpipe.cq`.
    bad_tokens : list of str
        The list of string tokens to remove from all token lists.
    result_columns : str or list-like, default None
        The names of the new columns resulting from the mapping operation.
        Must be of the same length as columns. If None, behavior depends on
        the drop parameter: If drop is True, the name of the source column
        is used; otherwise, the name of the source column is used with the
        suffix '_filtered'.
    drop : bool, default True
        If set to True, source columns are dropped after being transformed.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> data = [[4, ["a", "bad", "cat"]], [5, ["bad", "not", "good"]]]
    >>> df = pd.DataFrame(data, [1,2], ["age","text"])
    >>> filter_tokens = pdp.DropTokensByList('text', ['bad'])
    >>> filter_tokens(df)
       age         text
    1    4     [a, cat]
    2    5  [not, good]
    """  # noqa: W605

    class ListTokenFilter(object):

        def __init__(self, bad_tokens):
            self.bad_tokens = bad_tokens

        def __call__(self, token_list):
            return [x for x in token_list if x not in self.bad_tokens]

    def __init__(
        self, columns, bad_tokens, result_columns=None, drop=True,
        **kwargs
    ):
        self._bad_tokens = bad_tokens
        cond_str = ""
        if len(bad_tokens) < 10:
            cond_str = " in list [" + " ".join(bad_tokens) + "]"
        base_str = "Filtering out tokens{} in columns {{}}"
        desc_temp = base_str.format(cond_str)
        super_kwargs = {
            'columns': columns,
            'func': DropTokensByList.ListTokenFilter(bad_tokens),
            'result_columns': result_columns,
            'drop': drop,
            'suffix': "_filtered",
            'desc_temp': desc_temp,
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)



"""Defines the _BoundColumnPotential class."""

from types import MethodType
from typing import Union, Set, Dict, Tuple, Optional

import numpy
from pandas import DataFrame, Series

from ..shared import _list_str
from ..core import PdPipelineStage
from ..types import SeriesOperandTypesTuple
from .func_lists import (
    SERIES_TRANSFORMS_BLACKLIST,
    SERIES_TRANSFORMS_WHITELIST,
)
from .series_from_df import (
    _SeriesFromDf,
    _SeriesFromDfOperandType,
    _SeriesFromDfByLabel,
    _SeriesFromDfBySeriesMethod,
    _SeriesFromDfNeg,
    _SeriesFromDfAbs,
    _SeriesFromDfInvert,
    _SeriesFromDfAnd,
    _SeriesFromDfOr,
    _SeriesFromDfXor,
    _SeriesFromDfEq,
    _SeriesFromDfNe,
    _SeriesFromDfLt,
    _SeriesFromDfLe,
    _SeriesFromDfGt,
    _SeriesFromDfGe,
    _SeriesFromDfAdd,
    _SeriesFromDfSub,
    _SeriesFromDfMul,
    _SeriesFromDfTrueDiv,
    _SeriesFromDfMod,
    _SeriesFromDfPow,
    _SeriesFromDfFloorDiv,
)


class SeriesFromDfAssigner(PdPipelineStage):
    """Assigns a series to a column to input dataframes.

    Parameters
    ----------
    assign_to_column : str
        The label of the column to assign the series to. If it already exists
        in the input dataframe, it will be overwritten. Otherwise, it will be
        added to it.
    series_from_df : _SeriesFromDf or Series
        The df-to-series callable with which will be produced on stage
        application, or a series to assign.
    required_columns : Set of objects
        The labels of all columns required in input dataframes in order to
        calculate the series to assign to the destination column.
    source_column_potential : _BoundColumnPotential, optional
        The _BoundColumnPotential that `series_from_df` came from. Optional.
    """

    _DOC_TEMPLATE = "Assign column {} with {}"

    def __init__(
        self,
        assign_to_column: str,
        series_from_df: _SeriesFromDfOperandType,
        required_columns: Set[object],
        source_column_potential: Optional['_BoundColumnPotential'] = None,
        **kwargs,
    ) -> None:
        self.assign_to_column = assign_to_column
        self.series_from_df = series_from_df
        self.required_columns = required_columns
        self.source_column_potential = source_column_potential
        self._columns_str = _list_str(required_columns)
        exmsg = (
            f"Not all required columns {self._columns_str} "
            f"found in input dataframe!"
        )
        desc = SeriesFromDfAssigner._DOC_TEMPLATE.format(
            self.assign_to_column, self.series_from_df)
        super_kwargs = {
            'exmsg': exmsg,
            'desc': desc,
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, df: DataFrame) -> bool:
        return set(self.required_columns).issubset(df.columns)

    def _transform(self, df: DataFrame, verbose=None) -> DataFrame:
        return df.assign(**{self.assign_to_column: self.series_from_df})

    # === Binary Operators ===

    # --- Boolean Operators ---

    def __and__(self, other: object) -> 'SeriesFromDfAssigner':
        if isinstance(other, _BoundColumnPotential):
            if self.source_column_potential:
                new_col_potential = self.source_column_potential & other
                return SeriesFromDfAssigner(
                    assign_to_column=self.assign_to_column,
                    series_from_df=new_col_potential.series_from_df,
                    required_columns=new_col_potential.required_columns,
                    source_column_potential=new_col_potential,
                )
            # else, self.series_from_df is pandas.Series or scalar so
            # self.series_from_df & other should work
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfAnd(
                    other.series_from_df, self.series_from_df),
                required_columns=self.required_columns,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfAnd(self.series_from_df, other),
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __or__(self, other: object) -> 'SeriesFromDfAssigner':
        if isinstance(other, _BoundColumnPotential):
            if self.source_column_potential:
                new_col_potential = self.source_column_potential | other
                return SeriesFromDfAssigner(
                    assign_to_column=self.assign_to_column,
                    series_from_df=new_col_potential.series_from_df,
                    required_columns=new_col_potential.required_columns,
                    source_column_potential=new_col_potential,
                )
            # else, self.series_from_df is pandas.Series or scalar so
            # other.series_from_df | self.series_from_df should work
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfOr(
                    self.series_from_df, other.series_from_df),
                required_columns=self.required_columns,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfOr(self.series_from_df, other),
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __xor__(self, other: object) -> 'SeriesFromDfAssigner':
        if isinstance(other, _BoundColumnPotential):
            if self.source_column_potential:
                new_col_potential = self.source_column_potential ^ other
                return SeriesFromDfAssigner(
                    assign_to_column=self.assign_to_column,
                    series_from_df=new_col_potential.series_from_df,
                    required_columns=new_col_potential.required_columns,
                    source_column_potential=new_col_potential,
                )
            # else, self.series_from_df is pandas.Series or scalar so
            # self.series_from_df ^ other should work
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfXor(
                    self.series_from_df, other.series_from_df),
                required_columns=self.required_columns,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfXor(self.series_from_df, other),
                required_columns=self.required_columns,
            )
        return NotImplemented

    # --- Rich Comparison Operators ---

    def __lt__(self, other: object) -> 'SeriesFromDfAssigner':
        if isinstance(other, _BoundColumnPotential):
            if self.source_column_potential:
                new_col_potential = self.source_column_potential < other
                return SeriesFromDfAssigner(
                    assign_to_column=self.assign_to_column,
                    series_from_df=new_col_potential.series_from_df,
                    required_columns=new_col_potential.required_columns,
                    source_column_potential=new_col_potential,
                )
            # else, self.series_from_df is pandas.Series or scalar so
            # self.series_from_df < other should work
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfGt(
                    other.series_from_df, self.series_from_df),
                required_columns=self.required_columns,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfLt(self.series_from_df, other),
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __le__(self, other: object) -> 'SeriesFromDfAssigner':
        if isinstance(other, _BoundColumnPotential):
            if self.source_column_potential:
                new_col_potential = self.source_column_potential <= other
                return SeriesFromDfAssigner(
                    assign_to_column=self.assign_to_column,
                    series_from_df=new_col_potential.series_from_df,
                    required_columns=new_col_potential.required_columns,
                    source_column_potential=new_col_potential,
                )
            # else, self.series_from_df is pandas.Series or scalar so
            # self.series_from_df <= other should work
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfGe(
                    other.series_from_df, self.series_from_df),
                required_columns=self.required_columns,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfLe(self.series_from_df, other),
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __eq__(self, other: object) -> 'SeriesFromDfAssigner':
        if isinstance(other, _BoundColumnPotential):
            if self.source_column_potential:
                new_col_potential = self.source_column_potential == other
                return SeriesFromDfAssigner(
                    assign_to_column=self.assign_to_column,
                    series_from_df=new_col_potential.series_from_df,
                    required_columns=new_col_potential.required_columns,
                    source_column_potential=new_col_potential,
                )
            # else, self.series_from_df is pandas.Series or scalar so
            # self.series_from_df == other should work
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfEq(
                    other.series_from_df, self.series_from_df),
                required_columns=self.required_columns,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfEq(self.series_from_df, other),
                required_columns=self.required_columns,
            )
        return NotImplemented  # pragma: no cover

    def __ne__(self, other: object) -> 'SeriesFromDfAssigner':
        if isinstance(other, _BoundColumnPotential):
            if self.source_column_potential:
                new_col_potential = self.source_column_potential != other
                return SeriesFromDfAssigner(
                    assign_to_column=self.assign_to_column,
                    series_from_df=new_col_potential.series_from_df,
                    required_columns=new_col_potential.required_columns,
                    source_column_potential=new_col_potential,
                )
            # else, self.series_from_df is pandas.Series or scalar so
            # self.series_from_df != other should work
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfNe(
                    other.series_from_df, self.series_from_df),
                required_columns=self.required_columns,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfNe(self.series_from_df, other),
                required_columns=self.required_columns,
            )
        return NotImplemented  # pragma: no cover

    def __ge__(self, other: object) -> 'SeriesFromDfAssigner':
        if isinstance(other, _BoundColumnPotential):
            if self.source_column_potential:
                new_col_potential = self.source_column_potential >= other
                return SeriesFromDfAssigner(
                    assign_to_column=self.assign_to_column,
                    series_from_df=new_col_potential.series_from_df,
                    required_columns=new_col_potential.required_columns,
                    source_column_potential=new_col_potential,
                )
            # else, self.series_from_df is pandas.Series or scalar so
            # self.series_from_df >= other should work
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfLe(
                    other.series_from_df, self.series_from_df),
                required_columns=self.required_columns,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfGe(self.series_from_df, other),
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __gt__(self, other: object) -> 'SeriesFromDfAssigner':
        if isinstance(other, _BoundColumnPotential):
            if self.source_column_potential:
                new_col_potential = self.source_column_potential > other
                return SeriesFromDfAssigner(
                    assign_to_column=self.assign_to_column,
                    series_from_df=new_col_potential.series_from_df,
                    required_columns=new_col_potential.required_columns,
                    source_column_potential=new_col_potential,
                )
            # else, self.series_from_df is pandas.Series or scalar so
            # self.series_from_df > other should work
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfLt(
                    other.series_from_df, self.series_from_df),
                required_columns=self.required_columns,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfGt(self.series_from_df, other),
                required_columns=self.required_columns,
            )
        return NotImplemented

    # --- Arithmetic Operators ---

    def __add__(self, other: object) -> 'SeriesFromDfAssigner':
        if isinstance(other, _BoundColumnPotential):
            if self.source_column_potential:
                new_col_potential = self.source_column_potential + other
                return SeriesFromDfAssigner(
                    assign_to_column=self.assign_to_column,
                    series_from_df=new_col_potential.series_from_df,
                    required_columns=new_col_potential.required_columns,
                    source_column_potential=new_col_potential,
                )
            # else, self.series_from_df is pandas.Series or scalar so
            # self.series_from_df + other should work
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfAdd(
                    other.series_from_df, self.series_from_df),
                required_columns=self.required_columns,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfAdd(self.series_from_df, other),
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __radd__(self, other: object) -> 'SeriesFromDfAssigner':
        return self.__add__(other)  # pragma: no cover

    def __sub__(self, other: object) -> 'SeriesFromDfAssigner':
        if isinstance(other, _BoundColumnPotential):
            if self.source_column_potential:
                new_col_potential = self.source_column_potential - other
                return SeriesFromDfAssigner(
                    assign_to_column=self.assign_to_column,
                    series_from_df=new_col_potential.series_from_df,
                    required_columns=new_col_potential.required_columns,
                    source_column_potential=new_col_potential,
                )
            # else, self.series_from_df is pandas.Series or scalar so
            # self.series_from_df - other should work
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfSub(
                    self.series_from_df, other.series_from_df),
                required_columns=self.required_columns,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfSub(self.series_from_df, other),
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __mul__(self, other: object) -> 'SeriesFromDfAssigner':
        if isinstance(other, _BoundColumnPotential):
            if self.source_column_potential:
                new_col_potential = self.source_column_potential * other
                return SeriesFromDfAssigner(
                    assign_to_column=self.assign_to_column,
                    series_from_df=new_col_potential.series_from_df,
                    required_columns=new_col_potential.required_columns,
                    source_column_potential=new_col_potential,
                )
            # else, self.series_from_df is pandas.Series or scalar so
            # self.series_from_df * other should work
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfMul(
                    self.series_from_df, other.series_from_df),
                required_columns=self.required_columns,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfMul(self.series_from_df, other),
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __truediv__(self, other: object) -> 'SeriesFromDfAssigner':
        if isinstance(other, _BoundColumnPotential):
            if self.source_column_potential:
                new_col_potential = self.source_column_potential / other
                return SeriesFromDfAssigner(
                    assign_to_column=self.assign_to_column,
                    series_from_df=new_col_potential.series_from_df,
                    required_columns=new_col_potential.required_columns,
                    source_column_potential=new_col_potential,
                )
            # else, self.series_from_df is pandas.Series or scalar so
            # self.series_from_df / other should work
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfTrueDiv(
                    self.series_from_df, other.series_from_df),
                required_columns=self.required_columns,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfTrueDiv(
                    self.series_from_df, other),
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __floordiv__(self, other: object) -> 'SeriesFromDfAssigner':
        if isinstance(other, _BoundColumnPotential):
            if self.source_column_potential:
                new_col_potential = self.source_column_potential // other
                return SeriesFromDfAssigner(
                    assign_to_column=self.assign_to_column,
                    series_from_df=new_col_potential.series_from_df,
                    required_columns=new_col_potential.required_columns,
                    source_column_potential=new_col_potential,
                )
            # else, self.series_from_df is pandas.Series or scalar so
            # self.series_from_df // other should work
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfFloorDiv(
                    self.series_from_df, other.series_from_df),
                required_columns=self.required_columns,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfFloorDiv(
                    self.series_from_df, other),
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __mod__(self, other: object) -> 'SeriesFromDfAssigner':
        if isinstance(other, _BoundColumnPotential):
            if self.source_column_potential:
                new_col_potential = self.source_column_potential % other
                return SeriesFromDfAssigner(
                    assign_to_column=self.assign_to_column,
                    series_from_df=new_col_potential.series_from_df,
                    required_columns=new_col_potential.required_columns,
                    source_column_potential=new_col_potential,
                )
            # else, self.series_from_df is pandas.Series or scalar so
            # self.series_from_df % other should work
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfMod(
                    self.series_from_df, other.series_from_df),
                required_columns=self.required_columns,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfMod(self.series_from_df, other),
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __pow__(self, other: object) -> 'SeriesFromDfAssigner':
        if isinstance(other, _BoundColumnPotential):
            if self.source_column_potential:
                new_col_potential = self.source_column_potential ** other
                return SeriesFromDfAssigner(
                    assign_to_column=self.assign_to_column,
                    series_from_df=new_col_potential.series_from_df,
                    required_columns=new_col_potential.required_columns,
                    source_column_potential=new_col_potential,
                )
            # else, self.series_from_df is pandas.Series or scalar so
            # self.series_from_df ** other should work
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfPow(
                    self.series_from_df, other.series_from_df),
                required_columns=self.required_columns,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            return SeriesFromDfAssigner(
                assign_to_column=self.assign_to_column,
                series_from_df=_SeriesFromDfPow(self.series_from_df, other),
                required_columns=self.required_columns,
            )
        return NotImplemented


try:
    _BoundColumnPotentialOperandType = Union[
        '_BoundColumnPotential',
        Series,
        int,
        float,
        complex,
        int,
        bool,
        bytes,
        str,
        memoryview,
        numpy.int8,
        numpy.uint8,
        numpy.float16,
        numpy.timedelta64,
        numpy.object_,
        numpy.int16,
        numpy.uint16,
        numpy.float32,
        numpy.complex64,
        numpy.bytes_,
        numpy.int32,
        numpy.uint32,
        numpy.float64,
        numpy.complex128,
        numpy.str_,
        numpy.int64,
        numpy.uint64,
        numpy.float128,
        numpy.complex256,
        numpy.bool_,
        numpy.void,
        numpy.longlong,
        numpy.ulonglong,
        numpy.datetime64,
    ]
except AttributeError:  # pragma: no cover
    _BoundColumnPotentialOperandType = Union[
        '_BoundColumnPotential',
        Series,
        int,
        float,
        complex,
        int,
        bool,
        bytes,
        str,
        memoryview,
        numpy.int8,
        numpy.uint8,
        numpy.float16,
        numpy.timedelta64,
        numpy.object_,
        numpy.int16,
        numpy.uint16,
        numpy.float32,
        numpy.complex64,
        numpy.bytes_,
        numpy.int32,
        numpy.uint32,
        numpy.float64,
        numpy.str_,
        numpy.int64,
        numpy.uint64,
        numpy.bool_,
        numpy.void,
        numpy.longlong,
        numpy.ulonglong,
        numpy.datetime64,
    ]


class _BoundColumnPotential():

    def __init__(
        self,
        series_from_df: _SeriesFromDf,
        required_columns: Set[object],
    ) -> None:
        self.series_from_df = series_from_df
        self.required_columns = required_columns

    # === Unary operators ===

    def __neg__(self) -> '_BoundColumnPotential':
        return _BoundColumnPotential(
            series_from_df=_SeriesFromDfNeg(self.series_from_df),
            required_columns=self.required_columns,
        )

    def __pos__(self) -> '_BoundColumnPotential':
        return self

    def __abs__(self) -> '_BoundColumnPotential':
        return _BoundColumnPotential(
            series_from_df=_SeriesFromDfAbs(self.series_from_df),
            required_columns=self.required_columns,
        )

    def __invert__(self) -> '_BoundColumnPotential':
        return _BoundColumnPotential(
            series_from_df=_SeriesFromDfInvert(self.series_from_df),
            required_columns=self.required_columns,
        )

    # === Binary operators ===

    # --- Boolean Operators ---

    def __and__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        if isinstance(other, _BoundColumnPotential):
            series_from_df = _SeriesFromDfAnd(
                first=self.series_from_df,
                second=other.series_from_df,
            )
            req_cols = self.required_columns | other.required_columns
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=req_cols,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            series_from_df = _SeriesFromDfAnd(
                first=self.series_from_df,
                second=other,
            )
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __rand__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        return self & other  # pragma: no cover

    def __or__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        if isinstance(other, _BoundColumnPotential):
            series_from_df = _SeriesFromDfOr(
                first=self.series_from_df,
                second=other.series_from_df,
            )
            req_cols = self.required_columns | other.required_columns
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=req_cols,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            series_from_df = _SeriesFromDfOr(
                first=self.series_from_df,
                second=other,
            )
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __ror__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        return self | other  # pragma: no cover

    def __xor__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        if isinstance(other, _BoundColumnPotential):
            series_from_df = _SeriesFromDfXor(
                first=self.series_from_df,
                second=other.series_from_df,
            )
            req_cols = self.required_columns | other.required_columns
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=req_cols,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            series_from_df = _SeriesFromDfXor(
                first=self.series_from_df,
                second=other,
            )
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __rxor__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        return self ^ other  # pragma: no cover

    # --- Rich Comparison Operators ---

    def __lt__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        if isinstance(other, _BoundColumnPotential):
            series_from_df = _SeriesFromDfLt(
                first=self.series_from_df,
                second=other.series_from_df,
            )
            req_cols = self.required_columns | other.required_columns
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=req_cols,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            series_from_df = _SeriesFromDfLt(
                first=self.series_from_df,
                second=other,
            )
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __le__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        if isinstance(other, _BoundColumnPotential):
            series_from_df = _SeriesFromDfLe(
                first=self.series_from_df,
                second=other.series_from_df,
            )
            req_cols = self.required_columns | other.required_columns
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=req_cols,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            series_from_df = _SeriesFromDfLe(
                first=self.series_from_df,
                second=other,
            )
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __eq__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        if isinstance(other, _BoundColumnPotential):
            series_from_df = _SeriesFromDfEq(
                first=self.series_from_df,
                second=other.series_from_df,
            )
            req_cols = self.required_columns | other.required_columns
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=req_cols,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            series_from_df = _SeriesFromDfEq(
                first=self.series_from_df,
                second=other,
            )
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=self.required_columns,
            )
        return NotImplemented  # pragma: no cover

    def __ne__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        if isinstance(other, _BoundColumnPotential):
            series_from_df = _SeriesFromDfNe(
                first=self.series_from_df,
                second=other.series_from_df,
            )
            req_cols = self.required_columns | other.required_columns
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=req_cols,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            series_from_df = _SeriesFromDfNe(
                first=self.series_from_df,
                second=other,
            )
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=self.required_columns,
            )
        return NotImplemented  # pragma: no cover

    def __ge__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        if isinstance(other, _BoundColumnPotential):
            series_from_df = _SeriesFromDfGe(
                first=self.series_from_df,
                second=other.series_from_df,
            )
            req_cols = self.required_columns | other.required_columns
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=req_cols,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            series_from_df = _SeriesFromDfGe(
                first=self.series_from_df,
                second=other,
            )
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __gt__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        if isinstance(other, _BoundColumnPotential):
            series_from_df = _SeriesFromDfGt(
                first=self.series_from_df,
                second=other.series_from_df,
            )
            req_cols = self.required_columns | other.required_columns
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=req_cols,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            series_from_df = _SeriesFromDfGt(
                first=self.series_from_df,
                second=other,
            )
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=self.required_columns,
            )
        return NotImplemented

    # --- Arithmetic Operators ---

    def __add__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        if isinstance(other, _BoundColumnPotential):
            series_from_df = _SeriesFromDfAdd(
                first=self.series_from_df,
                second=other.series_from_df,
            )
            req_cols = self.required_columns | other.required_columns
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=req_cols,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            series_from_df = _SeriesFromDfAdd(
                first=self.series_from_df,
                second=other,
            )
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __radd__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        return self.__add__(other)

    def __sub__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        if isinstance(other, _BoundColumnPotential):
            series_from_df = _SeriesFromDfSub(
                first=self.series_from_df,
                second=other.series_from_df,
            )
            req_cols = self.required_columns | other.required_columns
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=req_cols,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            series_from_df = _SeriesFromDfSub(
                first=self.series_from_df,
                second=other,
            )
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __rsub__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        return self.__sub__(other)

    def __mul__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        if isinstance(other, _BoundColumnPotential):
            series_from_df = _SeriesFromDfMul(
                first=self.series_from_df,
                second=other.series_from_df,
            )
            req_cols = self.required_columns | other.required_columns
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=req_cols,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            series_from_df = _SeriesFromDfMul(
                first=self.series_from_df,
                second=other,
            )
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __rmul__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        return self.__mul__(other)

    def __truediv__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        if isinstance(other, _BoundColumnPotential):
            series_from_df = _SeriesFromDfTrueDiv(
                first=self.series_from_df,
                second=other.series_from_df,
            )
            req_cols = self.required_columns | other.required_columns
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=req_cols,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            series_from_df = _SeriesFromDfTrueDiv(
                first=self.series_from_df,
                second=other,
            )
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __rtruediv__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        return self.__truediv__(other)

    def __floordiv__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        if isinstance(other, _BoundColumnPotential):
            series_from_df = _SeriesFromDfFloorDiv(
                first=self.series_from_df,
                second=other.series_from_df,
            )
            req_cols = self.required_columns | other.required_columns
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=req_cols,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            series_from_df = _SeriesFromDfFloorDiv(
                first=self.series_from_df,
                second=other,
            )
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __rfloordiv__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        return self.__floordiv__(other)

    def __mod__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        if isinstance(other, _BoundColumnPotential):
            series_from_df = _SeriesFromDfMod(
                first=self.series_from_df,
                second=other.series_from_df,
            )
            req_cols = self.required_columns | other.required_columns
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=req_cols,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            series_from_df = _SeriesFromDfMod(
                first=self.series_from_df,
                second=other,
            )
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __rmod__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        return self.__mod__(other)

    def __pow__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        if isinstance(other, _BoundColumnPotential):
            series_from_df = _SeriesFromDfPow(
                first=self.series_from_df,
                second=other.series_from_df,
            )
            req_cols = self.required_columns | other.required_columns
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=req_cols,
            )
        if isinstance(other, SeriesOperandTypesTuple):
            series_from_df = _SeriesFromDfPow(
                first=self.series_from_df,
                second=other,
            )
            return _BoundColumnPotential(
                series_from_df=series_from_df,
                required_columns=self.required_columns,
            )
        return NotImplemented

    def __rpow__(
        self,
        other: _BoundColumnPotentialOperandType,
    ) -> '_BoundColumnPotential':
        return self.__pow__(other)

    # --- Our custom assignment operator ---

    def __lshift__(
        self,
        other: Union['_BoundColumnPotential', Series],
    ) -> PdPipelineStage:
        if isinstance(self.series_from_df, _SeriesFromDfByLabel):
            me: _SeriesFromDfByLabel = self.series_from_df
            try:
                return SeriesFromDfAssigner(
                    assign_to_column=me.column_label,
                    series_from_df=other.series_from_df,
                    required_columns=other.required_columns,
                    source_column_potential=other,
                )
            except AttributeError:  # other is Series
                return SeriesFromDfAssigner(
                    assign_to_column=me.column_label,
                    series_from_df=other,
                    required_columns=set(),
                )
        return NotImplemented  # pragma: no cover

    def __rshift__(
        self,
        other: Union['_BoundColumnPotential', Series],
    ) -> PdPipelineStage:
        if isinstance(other, _BoundColumnPotential):
            return other.__lshift__(self)
        # if isinstance(self.series_from_df, _SeriesFromDfByLabel):
        #     me: _SeriesFromDfByLabel = self.series_from_df
        #     try:
        #         return SeriesFromDfAssigner(
        #             assign_to_column=me.column_label,
        #             series_from_df=other.series_from_df,
        #             required_columns=other.required_columns,
        #             source_column_potential=other,
        #         )
        #     except AttributeError:  # other is Series
        #         return SeriesFromDfAssigner(
        #             assign_to_column=me.column_label,
        #             series_from_df=other,
        #             required_columns=set(),
        #         )
        return NotImplemented


# --- add series transformer methods to _BoundColumnPotential ---

class _BoundColumnPotentialSeriesMethodTransformerHandle():

    def __init__(self, method_name: str, doc: str) -> None:
        self._method_name = method_name
        self.__doc__ = doc

    @staticmethod
    def _cast_bound_col_potentials_to_series_from_df_objects(
        args: tuple, kwargs: dict,
    ) -> Tuple[Tuple, Dict, Set]:
        """Breaks down the input args tuple and kwargs dict,
        cast any included _BoundColumnPotential objects into _SeriesFromDf
        objects, and rebuilds the args tuple and kwargs dict, also constructing
        the the required column set, and returning all threeself.

        Returns
        -------
        args : tuple
            The args tuple with any _BoundColumnPotential objects replaced by
            _SeriesFromDf objects.
        kwargs : Dict
            The kwargs dict with any _BoundColumnPotential objects replaced by
            _SeriesFromDf objects.
        required_columns : set
            The set of all column labels required by the function.
        """
        required_columns = set()
        args_ = []
        for arg in args:
            if isinstance(arg, _BoundColumnPotential):
                args_.append(arg.series_from_df)
                required_columns |= arg.required_columns
            else:
                args_.append(arg)
        args = tuple(args_)
        kwargs_ = {}
        for k, v in kwargs.items():
            if isinstance(v, _BoundColumnPotential):
                kwargs_[k] = v.series_from_df
                required_columns |= v.required_columns
            else:
                kwargs_[k] = v
        kwargs = kwargs_
        return args, kwargs, required_columns

    def __call__(
        self,
        bound_col_instance: _BoundColumnPotential,
        *args,
        **kwargs: Dict[str, object],
    ) -> PdPipelineStage:
        res = self._cast_bound_col_potentials_to_series_from_df_objects(
            args=args,
            kwargs=kwargs,
        )
        args, kwargs, required_columns = res
        new_ser_from_df = _SeriesFromDfBySeriesMethod(
            source=bound_col_instance.series_from_df,
            method_name=self._method_name,
            args=args,
            kwargs=kwargs,
        )
        required_columns |= bound_col_instance.required_columns
        return _BoundColumnPotential(
            series_from_df=new_ser_from_df,
            required_columns=required_columns,
        )

    def __get__(self, instance, owner):
        return MethodType(self, instance) if instance else self


__RETURNS = 'Returns'
__SERIES = 'Series'


def _has_series_transform_doc(attr_name: str, attr: object) -> bool:
    if attr_name.startswith('_') or not callable(attr):
        return False
    try:
        doc_lines = attr.__doc__.split('\n')
        returns_line_index = None
        for i, line in enumerate(doc_lines):
            if __RETURNS in line:
                returns_line_index = i + 2
                break
        if returns_line_index:
            return_type_line = doc_lines[returns_line_index]
            if __SERIES in return_type_line:
                return True  # pragma: no cover
        return False
    except (AttributeError, IndexError):  # pragma: no cover
        return False


for attr_name in dir(Series):
    attr = getattr(Series, attr_name)
    add = False
    if attr_name in SERIES_TRANSFORMS_WHITELIST:
        add = True
    elif attr_name in SERIES_TRANSFORMS_BLACKLIST:
        add = False
    else:
        # check if documented return value points at a series-to-series func
        add = _has_series_transform_doc(attr_name, attr)
    if add:
        handle = _BoundColumnPotentialSeriesMethodTransformerHandle(
            method_name=attr_name,
            doc=attr.__doc__,
        )
        setattr(
            _BoundColumnPotential,
            attr_name,
            handle,
        )


# ==== factory method ====

def get_bound_column_potential_by_label(
    column_label: str,
) -> _BoundColumnPotential:
    """Return a bound column potential for a column with the given labelself.

    Parameters
    ----------
    column_label : str
            The label of the column to bind.

    Returns
    -------
    _BoundColumnPotential
            A bound column potential for the column with the given label.
    """
    return _BoundColumnPotential(
        series_from_df=_SeriesFromDfByLabel(column_label=column_label),
        required_columns={column_label},
    )



"""Define callables that generate pandas.Series from pandas.DataFrame."""

import abc
from typing import Union, Tuple

import numpy
from pandas import DataFrame, Series


# === Series from DataFrame objects ======

class _SeriesFromDf(abc.ABC):
    """A serializable callable that returns a pandas.Series from input
    pandas.DataFrame objects."""

    @abc.abstractmethod
    def __call__(self, df: DataFrame) -> Series:
        raise NotImplementedError


try:
    _SeriesFromDfOperandType = Union[
        _SeriesFromDf,
        Series,
        int,
        float,
        complex,
        int,
        bool,
        bytes,
        str,
        memoryview,
        numpy.int8,
        numpy.uint8,
        numpy.float16,
        numpy.timedelta64,
        numpy.object_,
        numpy.int16,
        numpy.uint16,
        numpy.float32,
        numpy.complex64,
        numpy.bytes_,
        numpy.int32,
        numpy.uint32,
        numpy.float64,
        numpy.complex128,
        numpy.str_,
        numpy.int64,
        numpy.uint64,
        numpy.float128,
        numpy.complex256,
        numpy.bool_,
        numpy.void,
        numpy.longlong,
        numpy.ulonglong,
        numpy.datetime64,
    ]
except AttributeError:  # pragma: no cover
    _SeriesFromDfOperandType = Union[
        _SeriesFromDf,
        Series,
        int,
        float,
        complex,
        int,
        bool,
        bytes,
        str,
        memoryview,
        numpy.int8,
        numpy.uint8,
        numpy.float16,
        numpy.timedelta64,
        numpy.object_,
        numpy.int16,
        numpy.uint16,
        numpy.float32,
        numpy.complex64,
        numpy.bytes_,
        numpy.int32,
        numpy.uint32,
        numpy.float64,
        numpy.str_,
        numpy.int64,
        numpy.uint64,
        numpy.bool_,
        numpy.void,
        numpy.longlong,
        numpy.ulonglong,
        numpy.datetime64,
    ]


class _SeriesFromDfByLabel(_SeriesFromDf):

    def __init__(self, column_label: object) -> None:
        self.column_label = column_label

    def __call__(self, df: DataFrame) -> Series:
        return df[self.column_label]

    def __repr__(self) -> str:
        return f"df[{self.column_label}]"


class _SeriesFromDfBySeriesMethod(_SeriesFromDf):

    def __init__(
        self,
        source: _SeriesFromDf,
        method_name: str,
        args: tuple,
        kwargs: dict,
    ) -> None:
        self.source = source
        self.method_name = method_name
        self.args = args
        self.kwargs = kwargs

    def __repr__(self) -> str:
        args_str = ""
        pos_args_str = ""
        if len(self.args) > 0:
            pos_args_str = ", ".join(map(str, self.args))
            args_str += pos_args_str
        kwargs_str = ""
        if len(self.kwargs) > 0:
            kwargs_str = ", ".join(
                f"{k}={v}" for k, v in self.kwargs.items())
            if len(self.args) > 0:
                args_str += ", "
            args_str += kwargs_str
        return (
            f"{self.source.__repr__()}.{self.method_name}({args_str})"
        )

    @staticmethod
    def _cast_series_from_df_to_series(
        args: tuple,
        kwargs: dict,
        df: DataFrame,
    ) -> Tuple[tuple, dict]:
        """Breaks down the input args tuple and kwargs dict,
        applies any _SeriesFromDf objects found to the input dataframe to

        Returns
        -------
        args : tuple
            The args tuple with any _SeriesFromDf objects replaced by the
            concrete Series objects.
        kwargs : Dict
            The kwargs dict with any _SeriesFromDf objects replaced by the
            concrete Series objects.
        """
        args_ = []
        for arg in args:
            if isinstance(arg, _SeriesFromDf):
                args_.append(arg(df))
            else:
                args_.append(arg)
        kwargs_ = {}
        for key, value in kwargs.items():
            if isinstance(value, _SeriesFromDf):
                kwargs_[key] = value(df)
            else:
                kwargs_[key] = value
        return tuple(args_), kwargs_

    def __call__(self, df: DataFrame) -> Series:
        source_res = self.source(df)
        method = getattr(source_res, self.method_name)
        args, kwargs = self._cast_series_from_df_to_series(
            self.args, self.kwargs, df)
        return method(*args, **kwargs)


# === repr help functions ======

def _arg_repr(arg: _SeriesFromDfOperandType) -> str:
    if isinstance(arg, Series):
        return "pandas.Series"
    return repr(arg)


# === Unary Operators ===

class _SeriesFromDfNeg(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
    ) -> None:
        self.first = first

    def __call__(self, df: DataFrame) -> Series:
        return - self.first(df)

    def __repr__(self) -> str:
        return f"-{_arg_repr(self.first)}"


# class _SeriesFromDfPos(_SeriesFromDf):
#
#     def __init__(
#         self,
#         first: _SeriesFromDf,
#     ) -> None:
#         self.first = first
#
#     def __call__(self, df: DataFrame) -> Series:
#         return + self.first(df)


class _SeriesFromDfAbs(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
    ) -> None:
        self.first = first

    def __call__(self, df: DataFrame) -> Series:
        return abs(self.first(df))

    def __repr__(self) -> str:
        return f"abs({_arg_repr(self.first)})"


class _SeriesFromDfInvert(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
    ) -> None:
        self.first = first

    def __call__(self, df: DataFrame) -> Series:
        return ~ self.first(df)

    def __repr__(self) -> str:
        return f"~{_arg_repr(self.first)}"


# === Binary Operators ===

# --- Boolean Operators ---

class _SeriesFromDfAnd(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
        second: _SeriesFromDfOperandType,
    ) -> None:
        self.first = first
        self.second = second

    def __call__(self, df: DataFrame) -> Series:
        try:
            return self.first(df) & self.second(df)
        except TypeError:
            return self.first(df) & self.second

    def __repr__(self) -> str:
        return f"{_arg_repr(self.first)} & {_arg_repr(self.second)}"


class _SeriesFromDfOr(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
        second: _SeriesFromDfOperandType,
    ) -> None:
        self.first = first
        self.second = second

    def __call__(self, df: DataFrame) -> Series:
        if callable(self.first):
            if callable(self.second):
                return self.first(df) | self.second(df)
            return self.first(df) | self.second
        if callable(self.second):
            return self.first | self.second(df)
        return self.first | self.second  # pragma: no cover

    def __repr__(self) -> str:
        return f"{_arg_repr(self.first)} | {_arg_repr(self.second)}"


class _SeriesFromDfXor(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
        second: _SeriesFromDfOperandType,
    ) -> None:
        self.first = first
        self.second = second

    def __call__(self, df: DataFrame) -> Series:
        if callable(self.first):
            if callable(self.second):
                return self.first(df) ^ self.second(df)
            return self.first(df) ^ self.second
        if callable(self.second):
            return self.first ^ self.second(df)
        return self.first ^ self.second  # pragma: no cover

    def __repr__(self) -> str:
        return f"{_arg_repr(self.first)} ^ {_arg_repr(self.second)}"


# --- Rich Comparison Operators ---

class _SeriesFromDfLt(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
        second: _SeriesFromDfOperandType,
    ) -> None:
        self.first = first
        self.second = second

    def __call__(self, df: DataFrame) -> Series:
        try:
            return self.first(df) < self.second(df)
        except TypeError:
            return self.first(df) < self.second

    def __repr__(self) -> str:
        return f"{_arg_repr(self.first)} < {_arg_repr(self.second)}"


class _SeriesFromDfLe(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
        second: _SeriesFromDfOperandType,
    ) -> None:
        self.first = first
        self.second = second

    def __call__(self, df: DataFrame) -> Series:
        try:
            return self.first(df) <= self.second(df)
        except TypeError:
            return self.first(df) <= self.second

    def __repr__(self) -> str:
        return f"{_arg_repr(self.first)} <= {_arg_repr(self.second)}"


class _SeriesFromDfEq(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
        second: _SeriesFromDfOperandType,
    ) -> None:
        self.first = first
        self.second = second

    def __call__(self, df: DataFrame) -> Series:
        try:
            return self.first(df) == self.second(df)
        except TypeError:
            return self.first(df) == self.second

    def __repr__(self) -> str:
        return f"{_arg_repr(self.first)} == {_arg_repr(self.second)}"


class _SeriesFromDfNe(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
        second: _SeriesFromDfOperandType,
    ) -> None:
        self.first = first
        self.second = second

    def __call__(self, df: DataFrame) -> Series:
        try:
            return self.first(df) != self.second(df)
        except TypeError:
            return self.first(df) != self.second

    def __repr__(self) -> str:
        return f"{_arg_repr(self.first)} != {_arg_repr(self.second)}"


class _SeriesFromDfGe(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
        second: _SeriesFromDfOperandType,
    ) -> None:
        self.first = first
        self.second = second

    def __call__(self, df: DataFrame) -> Series:
        try:
            return self.first(df) >= self.second(df)
        except TypeError:
            return self.first(df) >= self.second

    def __repr__(self) -> str:
        return f"{_arg_repr(self.first)} >= {_arg_repr(self.second)}"


class _SeriesFromDfGt(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
        second: _SeriesFromDfOperandType,
    ) -> None:
        self.first = first
        self.second = second

    def __call__(self, df: DataFrame) -> Series:
        try:
            return self.first(df) > self.second(df)
        except TypeError:
            return self.first(df) > self.second

    def __repr__(self) -> str:
        return f"{_arg_repr(self.first)} > {_arg_repr(self.second)}"


# --- Arithmetic Operators ---

class _SeriesFromDfAdd(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
        second: _SeriesFromDfOperandType,
    ) -> None:
        self.first = first
        self.second = second

    def __call__(self, df: DataFrame) -> Series:
        try:
            return self.first(df) + self.second(df)
        except TypeError:
            return self.first(df) + self.second

    def __repr__(self) -> str:
        return f"{_arg_repr(self.first)} + {_arg_repr(self.second)}"


class _SeriesFromDfSub(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
        second: _SeriesFromDfOperandType,
    ) -> None:
        self.first = first
        self.second = second

    def __call__(self, df: DataFrame) -> Series:
        if callable(self.first):
            if callable(self.second):
                return self.first(df) - self.second(df)
            return self.first(df) - self.second
        if callable(self.second):
            return self.first - self.second(df)
        return self.first - self.second  # pragma: no cover

    def __repr__(self) -> str:
        return f"{_arg_repr(self.first)} - {_arg_repr(self.second)}"


class _SeriesFromDfMul(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
        second: _SeriesFromDfOperandType,
    ) -> None:
        self.first = first
        self.second = second

    def __call__(self, df: DataFrame) -> Series:
        if callable(self.first):
            if callable(self.second):
                return self.first(df) * self.second(df)
            return self.first(df) * self.second
        if callable(self.second):
            return self.first * self.second(df)
        return self.first * self.second  # pragma: no cover

    def __repr__(self) -> str:
        return f"{_arg_repr(self.first)} * {_arg_repr(self.second)}"


class _SeriesFromDfTrueDiv(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
        second: _SeriesFromDfOperandType,
    ) -> None:
        self.first = first
        self.second = second

    def __call__(self, df: DataFrame) -> Series:
        if callable(self.first):
            if callable(self.second):
                return self.first(df) / self.second(df)
            return self.first(df) / self.second
        if callable(self.second):
            return self.first / self.second(df)
        return self.first / self.second  # pragma: no cover

    def __repr__(self) -> str:
        return f"{_arg_repr(self.first)} / {_arg_repr(self.second)}"


class _SeriesFromDfFloorDiv(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
        second: _SeriesFromDfOperandType,
    ) -> None:
        self.first = first
        self.second = second

    def __call__(self, df: DataFrame) -> Series:
        if callable(self.first):
            if callable(self.second):
                return self.first(df) // self.second(df)
            return self.first(df) // self.second
        if callable(self.second):
            return self.first // self.second(df)
        return self.first // self.second  # pragma: no cover

    def __repr__(self) -> str:
        return f"{_arg_repr(self.first)} // {_arg_repr(self.second)}"


class _SeriesFromDfMod(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
        second: _SeriesFromDfOperandType,
    ) -> None:
        self.first = first
        self.second = second

    def __call__(self, df: DataFrame) -> Series:
        if callable(self.first):
            if callable(self.second):
                return self.first(df) % self.second(df)
            return self.first(df) % self.second
        if callable(self.second):
            return self.first % self.second(df)
        return self.first % self.second  # pragma: no cover

    def __repr__(self) -> str:
        return f"{_arg_repr(self.first)} % {_arg_repr(self.second)}"


class _SeriesFromDfPow(_SeriesFromDf):

    def __init__(
        self,
        first: _SeriesFromDf,
        second: _SeriesFromDfOperandType,
    ) -> None:
        self.first = first
        self.second = second

    def __call__(self, df: DataFrame) -> Series:
        if callable(self.first):
            if callable(self.second):
                return self.first(df) ** self.second(df)
            return self.first(df) ** self.second
        if callable(self.second):
            return self.first ** self.second(df)
        return self.first ** self.second  # pragma: no cover

    def __repr__(self) -> str:
        return f"{_arg_repr(self.first)} ** {_arg_repr(self.second)}"



"""Test PdPipelineAndSklearnEstimator."""

from typing import Optional

import pytest
import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import GridSearchCV
from sklearn.utils.validation import check_is_fitted
from sklearn.exceptions import NotFittedError
from sklearn.metrics import fbeta_score, make_scorer

import pdpipe as pdp
from pdpipe.skintegrate import (
    PdPipelineAndSklearnEstimator,
    pdpipe_scorer_from_sklearn_scorer,
)


DF1 = pd.DataFrame(
    data=[
        [23, 'Jo', 'M', True, 0.07, 'USA', 'Living life to its fullest'],
        [52, 'Regina', 'F', False, 0.26, 'Germany', 'I hate cats'],
        [23, 'Dana', 'F', True, 0.3, 'USA', 'the pen is mightier then the sword'],  # noqa : E501
        [25, 'Bo', 'M', False, 2.3, 'Greece', 'all for one and one for all'],
        [80, 'Richy', 'M', False, 100.2, 'Finland', 'I gots the dollarz'],
        [60, 'Paul', 'M', True, 1.87, 'Denmark', 'blah'],
        [44, 'Derek', 'M', True, 1.1, 'Denmark', 'every life is precious'],
        [72, 'Regina', 'F', True, 7.1, 'Greece', 'all of you get off my porch'],  # noqa : E501
        [50, 'Jim', 'M', False, 0.2, 'Germany', 'boy do I love dogs and cats'],
        [80, 'Wealthus', 'F', False, 123.2, 'Finland', 'me likey them moniez'],
    ],
    columns=['Age', 'Name', 'Gender', 'Smoking', 'Savings', 'Country', 'Quote'],  # noqa : E501
)


class MyPipelineAndModel(PdPipelineAndSklearnEstimator):

    def __init__(
        self,
        savings_max_val: Optional[int] = 100,
        drop_gender: Optional[bool] = False,
        scale_numeric: Optional[bool] = False,
        ohencode_country: Optional[bool] = True,
        savings_bin_val: Optional[int] = None,
        fit_intercept: Optional[bool] = True,
    ):
        self.savings_max_val = savings_max_val
        self.drop_gender = drop_gender
        self.scale_numeric = scale_numeric
        self.ohencode_country = ohencode_country
        self.savings_bin_val = savings_bin_val
        self.fit_intercept = fit_intercept
        cols_to_drop = []
        stages = [
            pdp.ColDrop(['Name', 'Quote'], errors='ignore'),
        ]
        if savings_bin_val:
            stages.append(pdp.Bin({'Savings': [savings_bin_val]}, drop=False))
            stages.append(pdp.Encode('Savings_bin'))
        if scale_numeric:
            stages.append(pdp.Scale('MinMaxScaler'))
        if drop_gender:
            cols_to_drop.append('Gender')
        else:
            stages.append(pdp.Encode('Gender'))
        if ohencode_country:
            stages.append(pdp.OneHotEncode('Country'))
        else:
            cols_to_drop.append('Country')
        stages.append(pdp.ColDrop(cols_to_drop, errors='ignore'))
        pline = pdp.PdPipeline(stages)
        model = LogisticRegression(fit_intercept=fit_intercept)
        super().__init__(pipeline=pline, estimator=model)


@pytest.mark.skintegrate
def test_pdpipeline_and_sklearn_model():
    mp = MyPipelineAndModel(
        savings_max_val=101,
        drop_gender=True,
        scale_numeric=True,
        ohencode_country=True,
        savings_bin_val=1,
        fit_intercept=True,
    )
    assert isinstance(mp.pipeline, pdp.PdPipeline)
    assert isinstance(mp.estimator, LogisticRegression)
    assert callable(mp.score)
    assert isinstance(str(mp), str)
    assert isinstance(repr(mp), str)

    # X-y subsets
    df = DF1.copy()
    x_lbls = ['Age', 'Gender', 'Savings', 'Country']
    all_x = df[x_lbls]
    all_y = df['Smoking']

    # check inheritence of predict, predict_proba, etc.
    mp.fit(all_x, all_y)
    res = mp.predict(all_x)
    assert isinstance(res, np.ndarray)
    assert res.dtype == bool
    res = mp.predict_proba(all_x)
    assert isinstance(res, np.ndarray)
    assert res.dtype == float
    res = mp.predict_log_proba(all_x)
    assert isinstance(res, np.ndarray)
    assert res.dtype == float
    res = mp.decision_function(all_x)
    assert isinstance(res, np.ndarray)
    assert res.dtype == float
    res = mp.score(all_x, all_y)
    assert isinstance(res, float)
    assert isinstance(mp.classes_, np.ndarray)
    assert mp.classes_.dtype == bool

    gcv = GridSearchCV(
        estimator=mp,
        param_grid={
            'savings_max_val': [99, 101],
            'scale_numeric': [True, False],
        },
        cv=3,
    )
    with pytest.raises(NotFittedError):
        check_is_fitted(gcv)
    gcv.fit(all_x, all_y)
    assert check_is_fitted(gcv) is None
    assert isinstance(gcv.cv_results_, dict)
    assert isinstance(gcv.best_estimator_, PdPipelineAndSklearnEstimator)
    score1 = gcv.best_score_

    ftwo_scorer = make_scorer(fbeta_score, beta=2)
    my_scorer = pdpipe_scorer_from_sklearn_scorer(ftwo_scorer)
    assert isinstance(repr(my_scorer), str)
    gcv = GridSearchCV(
        estimator=mp,
        param_grid={
            'savings_max_val': [99, 101],
            'scale_numeric': [True, False],
        },
        cv=3,
        scoring=my_scorer,
    )
    gcv.fit(all_x, all_y)
    assert check_is_fitted(gcv) is None
    assert isinstance(gcv.cv_results_, dict)
    assert isinstance(gcv.best_estimator_, PdPipelineAndSklearnEstimator)
    score2 = gcv.best_score_
    assert score2 < score1


DF2 = pd.DataFrame(
    data=[['-1', 0], ['-1', 0], ['1', 1], ['1', 1]],
    index=[1, 2, 3, 4],
    columns=['feature1', 'target']
)


def test_pdpipeline_and_sklearn_model_documentation():
    all_x = DF2[['feature1']]
    all_y = DF2['target']
    mp = PdPipelineAndSklearnEstimator(
        pipeline=pdp.ColumnDtypeEnforcer({'feature1': int}),
        estimator=LogisticRegression()
    )
    mp.fit(all_x, all_y)
    res = mp.predict(all_x)
    assert isinstance(res, np.ndarray)
    assert len(res) == len(DF2)
    assert res.dtype == all_y.dtype



"""Test column bound potentials."""

import pytest
import pandas as pd
import numpy as np

import pdpipe as pdp
from pdpipe import df


def _num_df():
    return pd.DataFrame(
        data=[
            [-2, 2],
            [4, 3],
            [1, 1],
        ],
        columns=["a", "b"],
    )


def _pos_num_df():
    return pd.DataFrame(
        data=[
            [2, 2],
            [4, 3],
            [1, 1],
        ],
        columns=["a", "b"],
    )


def _bool_df():
    return pd.DataFrame(
        data=[
            [True, False],
            [True, True],
            [False, False],
            [False, True],
        ],
        columns=["a", "b"],
    )


def _num_df_with_na():
    return pd.DataFrame(
        data=[
            [-2, 2],
            [4, np.nan],
            [1, 1],
        ],
        columns=["a", "b"],
    )


class CustomType():
    """Custom type."""

    def __init__(self, a, b):
        self.a = a
        self.b = b


@pytest.mark.bound_col
def test_col_bound_potential_numerical_operators():
    """Test column bound potentials for numerical operators."""
    rdf = _num_df()
    c_vals = [8, 1, 3]
    d_val = 5
    # test unary operators
    pline = pdp.PdPipeline([
        df['-a'] << - df['a'],
        df['+a'] << + df['a'],
        df['abs(a)'] << abs(df['a']),
        df['c'] << pd.Series(c_vals),
        df['-c'] << -pd.Series(c_vals),
        df['d'] << d_val,
    ])
    res = pline(rdf)
    assert res['-a'].equals(-rdf['a'])
    assert res['+a'].equals(+rdf['a'])
    assert res['abs(a)'].equals(abs(rdf['a']))
    assert res['c'].equals(pd.Series(c_vals))
    assert res['-c'].equals(-pd.Series(c_vals))

    # test binary operators
    pline = pdp.PdPipeline([
        df['a<b'] << df['a'] < df['b'],
        df['a>b'] << df['a'] > df['b'],
        df['a<=b'] << df['a'] <= df['b'],
        df['a>=b'] << df['a'] >= df['b'],
        df['a==b'] << df['a'] == df['b'],
        df['a!=b'] << df['a'] != df['b'],
        df['a+b'] << df['a'] + df['b'],
        df['a-b'] << df['a'] - df['b'],
        df['a*b'] << df['a'] * df['b'],
        df['a/b'] << df['a'] / df['b'],
        df['a%b'] << df['a'] % df['b'],
        df['a**b'] << df['a'] ** df['b'],
        df['a//b'] << df['a'] // df['b'],
        # with c
        df['a<c'] << df['a'] < pd.Series(c_vals),
        df['a+d'] << df['a'] + d_val,
        # to the right
        (df['a'] + df['b']) >> df['a+b_right'],
    ])
    res = pline(rdf)
    assert res['a<b'].equals(rdf['a'] < rdf['b'])
    assert res['a>b'].equals(rdf['a'] > rdf['b'])
    assert res['a<=b'].equals(rdf['a'] <= rdf['b'])
    assert res['a>=b'].equals(rdf['a'] >= rdf['b'])
    assert res['a==b'].equals(rdf['a'].eq(rdf['b']))
    assert res['a!=b'].equals(rdf['a'] != rdf['b'])
    assert res['a+b'].equals(rdf['a'] + rdf['b'])
    assert res['a-b'].equals(rdf['a'] - rdf['b'])
    assert res['a*b'].equals(rdf['a'] * rdf['b'])
    assert res['a/b'].equals(rdf['a'] / rdf['b'])
    assert res['a%b'].equals(rdf['a'] % rdf['b'])
    assert res['a**b'].equals(rdf['a'] ** rdf['b'])
    assert res['a//b'].equals(rdf['a'] // rdf['b'])
    assert res['a<c'].equals(rdf['a'] < pd.Series(c_vals))
    assert res['a+d'].equals(rdf['a'] + d_val)
    assert res['a+b_right'].equals(rdf['a'] + rdf['b'])

    # test binary operators with SeriesFromDfAssigner
    pline = pdp.PdPipeline([
        (df['a+b'] << df['a']) + df['b'],
        (df['a-b'] << df['a']) - df['b'],
        (df['a*b'] << df['a']) * df['b'],
        (df['a/b'] << df['a']) / df['b'],
        (df['a%b'] << df['a']) % df['b'],
        (df['a**b'] << df['a']) ** df['b'],
        (df['a//b'] << df['a']) // df['b'],
        (df['a+d'] << df['a']) + d_val,
        df['a'] + (df['b'] >> df['a+b_right']),
    ])
    res = pline(rdf)
    assert res['a+b'].equals(rdf['a'] + rdf['b'])
    assert res['a-b'].equals(rdf['a'] - rdf['b'])
    assert res['a*b'].equals(rdf['a'] * rdf['b'])
    assert res['a/b'].equals(rdf['a'] / rdf['b'])
    assert res['a%b'].equals(rdf['a'] % rdf['b'])
    assert res['a**b'].equals(rdf['a'] ** rdf['b'])
    assert res['a//b'].equals(rdf['a'] // rdf['b'])
    assert res['a+d'].equals(rdf['a'] + d_val)
    assert res['a+b_right'].equals(rdf['a'] + rdf['b'])

    # test correct raising of TypeError
    custom = CustomType(1, 2)
    with pytest.raises(TypeError):
        df['a'] >> custom
    with pytest.raises(TypeError):
        df['a'] >> 4
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << (df['a'] < custom)])
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << (df['a'] > custom)])
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << (df['a'] <= custom)])
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << (df['a'] >= custom)])
    # with pytest.raises(TypeError):
    #     pdp.PdPipeline([df['c'] << (df['a'] == custom)])
    # with pytest.raises(TypeError):
    #     pdp.PdPipeline([df['c'] << (df['a'] != custom)])
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << (df['a'] + custom)])
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << (df['a'] - custom)])
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << (df['a'] * custom)])
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << (df['a'] / custom)])
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << (df['a'] % custom)])
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << (df['a'] ** custom)])
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << (df['a'] // custom)])

    # test correct raising of TypeError for SeriesFromDfAssigner
    with pytest.raises(TypeError):
        pdp.PdPipeline([(df['a+b'] << df['a']) < custom])
    with pytest.raises(TypeError):
        pdp.PdPipeline([(df['a+b'] << df['a']) > custom])
    with pytest.raises(TypeError):
        pdp.PdPipeline([(df['a+b'] << df['a']) <= custom])
    with pytest.raises(TypeError):
        pdp.PdPipeline([(df['a+b'] << df['a']) >= custom])
    # with pytest.raises(TypeError):
    #     pdp.PdPipeline([(df['a+b'] << df['a']) == custom])
    # with pytest.raises(TypeError):
    #     pdp.PdPipeline([(df['a+b'] << df['a']) != custom])
    with pytest.raises(TypeError):
        pdp.PdPipeline([(df['c'] << df['a']) + custom])
    with pytest.raises(TypeError):
        pdp.PdPipeline([(df['c'] << df['a']) - custom])
    with pytest.raises(TypeError):
        pdp.PdPipeline([(df['c'] << df['a']) * custom])
    with pytest.raises(TypeError):
        pdp.PdPipeline([(df['c'] << df['a']) / custom])
    with pytest.raises(TypeError):
        pdp.PdPipeline([(df['c'] << df['a']) % custom])
    with pytest.raises(TypeError):
        pdp.PdPipeline([(df['c'] << df['a']) ** custom])
    with pytest.raises(TypeError):
        pdp.PdPipeline([(df['c'] << df['a']) // custom])


@pytest.mark.bound_col
def test_col_bound_potential_numerical_operators_with_series():
    """Test column bound potentials for numerical operators."""
    rdf = _num_df()
    c_vals = [8, 1, 3]
    cs = pd.Series(c_vals)

    # test binary operators with series
    pline = pdp.PdPipeline([
        df['a<cs'] << (df['a'] < cs),
        df['a>cs'] << (df['a'] > cs),
        df['a<=cs'] << (df['a'] <= cs),
        df['a>=cs'] << (df['a'] >= cs),
        df['a==cs'] << (df['a'] == cs),
        df['a!=cs'] << (df['a'] != cs),
        df['a+cs'] << (df['a'] + cs),
        df['a-cs'] << (df['a'] - cs),
        df['a*cs'] << (df['a'] * cs),
        df['a/cs'] << (df['a'] / cs),
        df['a%cs'] << (df['a'] % cs),
        df['a**cs'] << (df['a'] ** cs),
        df['a//cs'] << (df['a'] // cs),
    ])
    res = pline(rdf)
    assert res['a<cs'].equals(rdf['a'] < cs)
    assert res['a>cs'].equals(rdf['a'] > cs)
    assert res['a<=cs'].equals(rdf['a'] <= cs)
    assert res['a>=cs'].equals(rdf['a'] >= cs)
    assert res['a==cs'].equals(rdf['a'].eq(cs))
    assert res['a!=cs'].equals(rdf['a'] != cs)
    assert res['a+cs'].equals(rdf['a'] + cs)
    assert res['a-cs'].equals(rdf['a'] - cs)
    assert res['a*cs'].equals(rdf['a'] * cs)
    assert res['a/cs'].equals(rdf['a'] / cs)
    assert res['a%cs'].equals(rdf['a'] % cs)
    assert res['a**cs'].equals(rdf['a'] ** cs)
    assert res['a//cs'].equals(rdf['a'] // cs)

    # test binary operators with series for SeriesFromDfAssigner
    pline = pdp.PdPipeline([
        (df['a<cs'] << df['a']) < cs,
        (df['a>cs'] << df['a']) > cs,
        (df['a<=cs'] << df['a']) <= cs,
        (df['a>=cs'] << df['a']) >= cs,
        (df['a==cs'] << df['a']) == cs,
        (df['a!=cs'] << df['a']) != cs,
        (df['a+cs'] << df['a']) + cs,
        (df['a-cs'] << df['a']) - cs,
        (df['a*cs'] << df['a']) * cs,
        (df['a/cs'] << df['a']) / cs,
        (df['a%cs'] << df['a']) % cs,
        (df['a**cs'] << df['a']) ** cs,
        (df['a//cs'] << df['a']) // cs,
    ])
    res = pline(rdf)
    assert res['a<cs'].equals(rdf['a'] < cs)
    assert res['a>cs'].equals(rdf['a'] > cs)
    assert res['a<=cs'].equals(rdf['a'] <= cs)
    assert res['a>=cs'].equals(rdf['a'] >= cs)
    assert res['a==cs'].equals(rdf['a'].eq(cs))
    assert res['a!=cs'].equals(rdf['a'] != cs)
    assert res['a+cs'].equals(rdf['a'] + cs)
    assert res['a-cs'].equals(rdf['a'] - cs)
    assert res['a*cs'].equals(rdf['a'] * cs)
    assert res['a/cs'].equals(rdf['a'] / cs)
    assert res['a%cs'].equals(rdf['a'] % cs)
    assert res['a**cs'].equals(rdf['a'] ** cs)
    assert res['a//cs'].equals(rdf['a'] // cs)

    # test binary operators with series on the left
    pline = pdp.PdPipeline([
        df['cs<a'] << cs < df['a'],
        df['cs>a'] << cs > df['a'],
        df['cs<=a'] << cs <= df['a'],
        df['cs>=a'] << cs >= df['a'],
        df['cs==a'] << cs == df['a'],
        df['cs!=a'] << cs != df['a'],
        # these cases are not supported, since operator resolution order means
        # that the Series object on the left is enquired first for the
        # support of the __add__, __sub__, etc. methods for a
        # _BoundColumnPotential object, which they always support, interpreting
        # it as an object to add/sub/etc. element-wise. As such, the __radd__,
        # __rsub__, etc. methods of the _BoundColumnPotential object on the
        # right are never called, and we can't support the operation.
        # df['cs+a'] << cs + df['a'],
        # df['cs-a'] << cs - df['a'],
        # df['cs*a'] << cs * df['a'],
        # df['cs/a'] << cs / df['a'],
        # df['cs%a'] << cs % df['a'],
        # df['cs**a'] << cs ** df['a'],
        # df['cs//a'] << cs // df['a'],
    ])
    res = pline(rdf)
    assert res['cs<a'].equals(cs < rdf['a'])
    assert res['cs>a'].equals(cs > rdf['a'])
    assert res['cs<=a'].equals(cs <= rdf['a'])
    assert res['cs>=a'].equals(cs >= rdf['a'])
    assert res['cs==a'].equals(cs.eq(rdf['a']))
    assert res['cs!=a'].equals(cs != rdf['a'])
    # assert res['cs+a'].equals(cs + rdf['a'])
    # assert res['cs-a'].equals(cs - rdf['a'])
    # assert res['cs*a'].equals(cs * rdf['a'])
    # assert res['cs/a'].equals(cs / rdf['a'])
    # assert res['cs%a'].equals(cs % rdf['a'])
    # assert res['cs**a'].equals(cs ** rdf['a'])
    # assert res['cs//a'].equals(cs // rdf['a'])

    # test binary operators with a scalar on the left
    cs = 8
    pline = pdp.PdPipeline([
        df['cs<a'] << cs < df['a'],
        df['cs>a'] << cs > df['a'],
        df['cs<=a'] << cs <= df['a'],
        df['cs>=a'] << cs >= df['a'],
        df['cs==a'] << cs == df['a'],
        df['cs!=a'] << cs != df['a'],
        df['cs+a'] << cs + df['a'],
        df['cs-a'] << cs - df['a'],
        df['cs*a'] << cs * df['a'],
        df['cs/a'] << cs / df['a'],
        df['cs%a'] << cs % df['a'],
        df['cs**a'] << cs ** df['a'],
        df['cs//a'] << cs // df['a'],
    ])
    res = pline(rdf)
    assert res['cs<a'].equals(cs < rdf['a'])
    assert res['cs>a'].equals(cs > rdf['a'])
    assert res['cs<=a'].equals(cs <= rdf['a'])
    assert res['cs>=a'].equals(cs >= rdf['a'])
    assert res['cs==a'].equals(cs == rdf['a'])
    assert res['cs!=a'].equals(cs != rdf['a'])
    assert res['cs+a'].equals(cs + rdf['a'])
    # assert res['cs-a'].equals(cs - rdf['a'])
    # assert res['cs*a'].equals(cs * rdf['a'])
    # assert res['cs/a'].equals(cs / rdf['a'])
    # assert res['cs%a'].equals(cs % rdf['a'])

    # test binary operators with a scalar on the left
    rdf = _pos_num_df()
    cs = 8
    pline = pdp.PdPipeline([
        (df['cs+a'] << cs) + df['a'],
        (df['cs-a'] << cs) - df['a'],
        (df['cs*a'] << cs) * df['a'],
        (df['cs/a'] << cs) / df['a'],
        (df['cs%a'] << cs) % df['a'],
        (df['cs**a'] << cs) ** df['a'],
        (df['cs//a'] << cs) // df['a'],
    ])
    res = pline(rdf)
    assert res['cs+a'].equals(cs + rdf['a'])
    assert res['cs-a'].equals(cs - rdf['a'])
    assert res['cs*a'].equals(cs * rdf['a'])
    assert res['cs/a'].equals(cs / rdf['a'])
    assert res['cs%a'].equals(cs % rdf['a'])


@pytest.mark.bound_col
def test_col_bound_potential_boolean_operators():
    """Test column bound potentials for boolean operators."""
    rdf = _bool_df()
    # test unary operators
    pline = pdp.PdPipeline([
        df['not(a)'] << ~df['a'],
    ])
    res = pline(rdf)
    assert res['not(a)'].equals(~rdf['a'])

    # test binary operators
    pline = pdp.PdPipeline([
        df['a&b'] << df['a'] & df['b'],
        df['a|b'] << df['a'] | df['b'],
        df['a^b'] << df['a'] ^ df['b'],
        df['a&~b'] << df['a'] & ~df['b'],
        df['a|~b'] << df['a'] | ~df['b'],
        df['a^~b'] << df['a'] ^ ~df['b'],
    ])
    res = pline(rdf)
    assert res['a&b'].equals(rdf['a'] & rdf['b'])
    assert res['a|b'].equals(rdf['a'] | rdf['b'])
    assert res['a^b'].equals(rdf['a'] ^ rdf['b'])
    assert res['a&~b'].equals(rdf['a'] & ~rdf['b'])
    assert res['a|~b'].equals(rdf['a'] | ~rdf['b'])
    assert res['a^~b'].equals(rdf['a'] ^ ~rdf['b'])

    # test binary operators with series
    bool_series = pd.Series([True, False, True])
    pline = pdp.PdPipeline([
        df['a&b'] << df['a'] & bool_series,
        df['a|b'] << df['a'] | bool_series,
        df['a^b'] << df['a'] ^ bool_series,
        df['a&~b'] << df['a'] & ~bool_series,
        df['a|~b'] << df['a'] | ~bool_series,
        df['a^~b'] << df['a'] ^ ~bool_series,
    ])
    res = pline(rdf)
    assert res['a&b'].equals(rdf['a'] & bool_series)
    assert res['a|b'].equals(rdf['a'] | bool_series)
    assert res['a^b'].equals(rdf['a'] ^ bool_series)
    assert res['a&~b'].equals(rdf['a'] & ~bool_series)
    assert res['a|~b'].equals(rdf['a'] | ~bool_series)
    assert res['a^~b'].equals(rdf['a'] ^ ~bool_series)

    # test binary operators with series
    bool_series = pd.Series([True, False, True])
    pline = pdp.PdPipeline([
        df['a&b'] << (df['a'] & bool_series),
        df['a|b'] << (df['a'] | bool_series),
        df['a^b'] << (df['a'] ^ bool_series),
        df['a&~b'] << (df['a'] & ~bool_series),
        df['a|~b'] << (df['a'] | ~bool_series),
        df['a^~b'] << (df['a'] ^ ~bool_series),
    ])
    res = pline(rdf)
    assert res['a&b'].equals(rdf['a'] & bool_series)
    assert res['a|b'].equals(rdf['a'] | bool_series)
    assert res['a^b'].equals(rdf['a'] ^ bool_series)
    assert res['a&~b'].equals(rdf['a'] & ~bool_series)
    assert res['a|~b'].equals(rdf['a'] | ~bool_series)
    assert res['a^~b'].equals(rdf['a'] ^ ~bool_series)

    # test binary operators with seris on the left
    # === can't be supported for the same reason - pd.Series hijacks the binary
    # operator when on the left
    # maybe check this out one day:
    # https://pandas.pydata.org/docs/development/extending.html#extensionarray-operator-support
    # bool_series = pd.Series([True, False, True])
    # pline = pdp.PdPipeline([
    #     df['a&b'] << (bool_series & df['b']),
    #     df['a|b'] << (bool_series | df['b']),
    #     df['a^b'] << (bool_series ^ df['b']),
    #     df['a&~b'] << (bool_series & ~df['b']),
    #     df['a|~b'] << (bool_series | ~df['b']),
    #     df['a^~b'] << (bool_series ^ ~df['b']),
    # ])
    # res = pline(rdf)
    # assert res['a&b'].equals(bool_series & rdf['b'])
    # assert res['a|b'].equals(bool_series | rdf['b'])
    # assert res['a^b'].equals(bool_series ^ rdf['b'])
    # assert res['a&~b'].equals(bool_series & ~rdf['b'])
    # assert res['a|~b'].equals(bool_series | ~rdf['b'])
    # assert res['a^~b'].equals(bool_series ^ ~rdf['b'])

    # test binary operators with seris on the left for SeriesFromDfAssigner
    bool_series = pd.Series([True, False, True])
    pline = pdp.PdPipeline([
        df['a&b'] << bool_series & df['b'],
        df['a|b'] << bool_series | df['b'],
        df['a^b'] << bool_series ^ df['b'],
        df['a&~b'] << bool_series & ~df['b'],
        df['a|~b'] << bool_series | ~df['b'],
        df['a^~b'] << bool_series ^ ~df['b'],
    ])
    res = pline(rdf)
    assert res['a&b'].equals(bool_series & rdf['b'])
    assert res['a|b'].equals(bool_series | rdf['b'])
    assert res['a^b'].equals(bool_series ^ rdf['b'])
    assert res['a&~b'].equals(bool_series & ~rdf['b'])
    assert res['a|~b'].equals(bool_series | ~rdf['b'])
    assert res['a^~b'].equals(bool_series ^ ~rdf['b'])

    # test correct raising of TypeError
    custom = CustomType(1, 2)
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << (df['a'] & custom)])
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << (df['a'] | custom)])
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << (df['a'] ^ custom)])

    # test correct raising of TypeError for SeriesFromDfAssigner
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << df['a'] & custom])
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << df['a'] | custom])
    with pytest.raises(TypeError):
        pdp.PdPipeline([df['c'] << df['a'] ^ custom])


@pytest.mark.bound_col
def test_col_bound_potential_complex():
    """Test column bound potentials for boolean operators."""
    rdf = _num_df_with_na()
    c_vals = [8, 1, 3]
    c_s = pd.Series(c_vals)
    d_val = 5
    # test unary operators
    pline = pdp.PdPipeline([
        df['-a'] << - df['a'].isin([3, 4, 5]),
        df['filled_b'] << df['b'].fillna(0),
    ])
    res = pline(rdf)
    assert res['-a'].equals(-rdf['a'].isin([3, 4, 5]))
    assert res['filled_b'].equals(rdf['b'].fillna(0))

    # test binary operators
    MAP = {-2: 7, 1: 5, 4: 1}
    pline = pdp.PdPipeline([
        df['a_map'] << df['a'].map(MAP),
        df['a<b'] << df['a'].map(MAP) < df['b'].fillna(1),
        df['a_map>b+d'] << df['a'].map(MAP) > df['b'] + d_val,
        df['a_map+d'] << df['a'].map(MAP) + d_val,
        df['a_map>d+b'] << df['a'].map(MAP) > d_val + df['b'],
        df['a_map+b+c'] << df['a'].map(MAP) + df['b'] + c_s,
        df['a_add_b'] << df['a'].add(df['b'], fill_value=0),
        df['a_add_b_kwarg'] << df['a'].add(other=df['b'], fill_value=0),
    ])
    res = pline(rdf)
    assert res['a_map'].equals(rdf['a'].map(MAP))
    assert res['a<b'].equals(rdf['a'].map(MAP) < rdf['b'].fillna(1))
    assert res['a_map>b+d'].equals(rdf['a'].map(MAP) > rdf['b'] + d_val)
    assert res['a_map+d'].equals(rdf['a'].map(MAP) + d_val)
    assert res['a_map>d+b'].equals(rdf['a'].map(MAP) > d_val + rdf['b'])
    assert res['a_map+b+c'].equals(rdf['a'].map(MAP) + rdf['b'] + c_s)
    assert res['a_add_b'].equals(rdf['a'].add(rdf['b'], fill_value=0))
    assert res['a_add_b_kwarg'].equals(
        rdf['a'].add(other=rdf['b'], fill_value=0))



"""Testing sklearn_stages.LabelEncoder."""

import math

import pytest
import numpy as np
import pandas as pd

import pdpipe as pdp
from pdpipe.sklearn_stages import EncodeLabel
from pdpipe.exceptions import UnfittedPipelineStageError


def _some_X_y():
    X = pd.DataFrame(
        data=[[3.2, 31], [7.2, 33], [12.1, 28]],
        index=[1, 2, 3],
        columns=["ph", "temp"]
    )
    y = pd.Series(["acd", "alk", "alk"])
    return X, y


def _some_X_y2():
    X = pd.DataFrame(
        data=[[4.1, 22], [7.7, 23], [2.1, 38]],
        index=[1, 2, 3],
        columns=["ph", "temp"]
    )
    y = pd.Series(["alk", "acd", "alk"])
    return X, y


def test_encode_label():
    X, y = _some_X_y()
    encode_stage = EncodeLabel()

    post_X, post_y = encode_stage(X, y)
    assert ['ph', 'temp'] == list(post_X.columns)
    assert post_y[1] == 0
    assert post_y[2] == 1
    assert post_y[3] == 1

    # see only transform (no fit) when already fitted
    X2, y2 = _some_X_y2()
    post_X, post_y = encode_stage(X2, y2)
    assert ['ph', 'temp'] == list(post_X.columns)
    assert post_y[1] == 1
    assert post_y[2] == 0
    assert post_y[3] == 1

    # see only transform (no fit) when already fitted
    X2, y2 = _some_X_y2()
    post_X, post_y = encode_stage.transform(X2, y2)
    assert ['ph', 'temp'] == list(post_X.columns)
    assert post_y[1] == 1
    assert post_y[2] == 0
    assert post_y[3] == 1

    # check fit_transform when already fitted
    X2, y2 = _some_X_y2()
    post_X, post_y = encode_stage.fit_transform(X2, y2)
    assert ['ph', 'temp'] == list(post_X.columns)
    assert post_y[1] == 1
    assert post_y[2] == 0
    assert post_y[3] == 1


def test_encode_label_skip():
    X, y = _some_X_y()
    pline = pdp.PdPipeline([
        pdp.ApplyByCols("ph", math.ceil),
        pdp.EncodeLabel(skip=lambda X, y: 'temp' in X.columns),
    ])

    post_X, post_y = pline(X, y)
    assert ['ph', 'temp'] == list(post_X.columns)
    assert post_y[1] == 'acd'
    assert post_y[2] == 'alk'
    assert post_y[3] == 'alk'


def test_encode_label_fit():
    X, y = _some_X_y()
    encode_stage = EncodeLabel()

    with pytest.raises(UnfittedPipelineStageError):
        encode_stage.transform(X, y)

    encode_stage.fit(X, y)
    post_X, post_y = encode_stage.transform(X, y)
    assert ['ph', 'temp'] == list(post_X.columns)
    assert post_y[1] == 0
    assert post_y[2] == 1
    assert post_y[3] == 1

    # see only transform (no fit) when already fitted
    X2, y2 = _some_X_y2()
    post_X, post_y = encode_stage.transform(X2, y2)
    assert ['ph', 'temp'] == list(post_X.columns)
    assert post_y[1] == 1
    assert post_y[2] == 0
    assert post_y[3] == 1


def _bad_len_X_y():
    X = pd.DataFrame(
        data=[[3.2, 31], [7.2, 33], [12.1, 28]],
        index=[1, 2, 3],
        columns=["ph", "temp"]
    )
    y = pd.Series(["acd", "alk"])
    return X, y


def test_encode_label_bad_len():
    X, y = _bad_len_X_y()
    encode_stage = EncodeLabel()

    with pytest.raises(ValueError):
        encode_stage.fit(X, y)

    with pytest.raises(ValueError):
        encode_stage(X, y)


def _np_X_y():
    X = pd.DataFrame(
        data=[[3.2, 31], [7.2, 33], [12.1, 28]],
        index=[1, 2, 3],
        columns=["ph", "temp"]
    )
    y = np.array(["acd", "alk", "alk"])
    return X, y


def _np_X_y2():
    X = pd.DataFrame(
        data=[[4.1, 22], [7.7, 23], [2.1, 38]],
        index=[1, 2, 3],
        columns=["ph", "temp"]
    )
    y = np.array(["alk", "acd", "alk"])
    return X, y


def test_encode_label_np_arr():
    X, y = _np_X_y()
    encode_stage = EncodeLabel()

    post_X, post_y = encode_stage(X, y)
    assert ['ph', 'temp'] == list(post_X.columns)
    assert post_y[1] == 0
    assert post_y[2] == 1
    assert post_y[3] == 1

    # see only transform (no fit) when already fitted
    X2, y2 = _np_X_y2()
    post_X, post_y = encode_stage(X2, y2)
    assert ['ph', 'temp'] == list(post_X.columns)
    assert post_y[1] == 1
    assert post_y[2] == 0
    assert post_y[3] == 1

    # see only transform (no fit) when already fitted
    X2, y2 = _np_X_y2()
    post_X, post_y = encode_stage.transform(X2, y2)
    assert ['ph', 'temp'] == list(post_X.columns)
    assert post_y[1] == 1
    assert post_y[2] == 0
    assert post_y[3] == 1

    # check fit_transform when already fitted
    X2, y2 = _np_X_y2()
    post_X, post_y = encode_stage.fit_transform(X2, y2)
    assert ['ph', 'temp'] == list(post_X.columns)
    assert post_y[1] == 1
    assert post_y[2] == 0
    assert post_y[3] == 1



"""On the fly, relative pipeline stage creation.

Use `drop_rows_where` and `keep_rows_where` as handles to the future dataframe,
using the `[]` indexing syntax to select a (single) column to apply the logic
by, and regular binary operators such as >, >=, ==, !=, etc. to express the
condition by which rows will be kept or dropped.

For example:

```python
>>> import pandas as pd; import pdpipe as pdp;
>>> df = pd.DataFrame([[1,4],[4,5],[5,11]], [1,2,3], ['a','b'])
>>> df
   a   b
1  1   4
2  4   5
3  5  11
>>> pipeline = pdp.PdPipeline([
...     pdp.drop_rows_where['a'] > 4,
... ])
>>> pipeline(df)
   a  b
1  1  4
2  4  5

```

The resulting stages can be naturaly combined by logical binary operators:
& for AND, | for OR and ^ for XOR, and can also be inverted with the `~`
operator.

For example:

```python
>>> import pandas as pd; import pdpipe as pdp;
>>> df = pd.DataFrame([[1,4],[4,5],[5,11]], [1,2,3], ['a','b'])
>>> pipeline = pdp.PdPipeline([
...     ~ (pdp.drop_rows_where['a'] > 4),
... ])
>>> pipeline(df)
   a   b
3  5  11
>>> pipeline = pdp.PdPipeline([
...     (pdp.drop_rows_where['a'] > 3) & (pdp.drop_rows_where['b'] < 10),
... ])
>>> pipeline(df)
   a   b
1  1   4
3  5  11

```
"""

from typing import List, Set, Union

import pandas

from .core import PdPipelineStage
from . import rq


# === Auxilary pipeline stages ===

class KeepRowsByQualifier(PdPipelineStage):
    """A pipeline stage that keeps rows by a row qualifier.

    All rows which the qualifier qualifies (i.e. return a boolean series with
    True in the corresponding entries) will be kept, while all other rows will
    be dropped from input dataframes.

    Parameters
    ----------
    qualifier : RowQualifier
        An object that returns a boolean series from input dataframes. See more
        in `pdpipe.rq`.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[1,4],[4,5],[5,11]], [1,2,3], ['a','b'])
    >>> q = pdp.rq.ColValGt('a', 3)
    >>> pdp.fly.KeepRowsByQualifier(q).apply(df)
       a   b
    2  4   5
    3  5  11
    """

    def __init__(self, qualifier, **kwargs):
        self._keeprowsby_rq = qualifier
        super_kwargs = {
            'desc': f'Drop rows by qualifier {qualifier}',
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, X: pandas.DataFrame) -> bool:
        return True

    def _transform(self, X, verbose=None):
        before_count = len(X)
        bool_ix = self._keeprowsby_rq(X)
        inter_X = X[bool_ix]
        if verbose:
            print(f"{before_count - len(inter_X)} rows dropped.")
        return inter_X

    def __and__(self, other):
        try:
            and_rq = self._keeprowsby_rq & other._keeprowsby_rq
            return type(self)(qualifier=and_rq)
        except AttributeError:
            return NotImplemented

    def __or__(self, other):
        try:
            or_rq = self._keeprowsby_rq | other._keeprowsby_rq
            return type(self)(qualifier=or_rq)
        except AttributeError:
            return NotImplemented

    def __xor__(self, other):
        try:
            xor_rq = self._keeprowsby_rq ^ other._keeprowsby_rq
            return type(self)(qualifier=xor_rq)
        except AttributeError:
            return NotImplemented

    def __invert__(self):
        not_rq = ~ self._keeprowsby_rq
        return type(self)(qualifier=not_rq)


class DropRowsByQualifier(PdPipelineStage):
    """A pipeline stage that drops rows by a row qualifier.

    All rows which the qualifier qualifies (i.e. return a boolean series with
    True in the corresponding entries) will be dropped, while all other rows
    will be kept in input dataframes.

    Parameters
    ----------
    qualifier : RowQualifier
        An object that returns a boolean series from input dataframes. See more
        in `pdpipe.rq`.

    Example
    -------
    >>> import pandas as pd; import pdpipe as pdp;
    >>> df = pd.DataFrame([[1,4],[4,5],[5,11]], [1,2,3], ['a','b'])
    >>> q = pdp.rq.ColValLt('a', 3)
    >>> pdp.fly.DropRowsByQualifier(q).apply(df)
       a   b
    2  4   5
    3  5  11
    """

    def __init__(self, qualifier, **kwargs):
        self._droprowsby_rq = qualifier
        super_kwargs = {
            'desc': f'Drop rows by qualifier {qualifier}',
        }
        super_kwargs.update(**kwargs)
        super().__init__(**super_kwargs)

    def _prec(self, X: pandas.DataFrame) -> bool:
        return True

    def _transform(self, X, verbose=None):
        before_count = len(X)
        bool_ix = ~ self._droprowsby_rq(X)
        inter_X = X[bool_ix]
        if verbose:
            print(f"{before_count - len(inter_X)} rows dropped.")
        return inter_X

    def __and__(self, other):
        try:
            and_rq = self._droprowsby_rq & other._droprowsby_rq
            return type(self)(qualifier=and_rq)
        except AttributeError:
            return NotImplemented

    def __or__(self, other):
        try:
            or_rq = self._droprowsby_rq | other._droprowsby_rq
            return type(self)(qualifier=or_rq)
        except AttributeError:
            return NotImplemented

    def __xor__(self, other):
        try:
            xor_rq = self._droprowsby_rq ^ other._droprowsby_rq
            return type(self)(qualifier=xor_rq)
        except AttributeError:
            return NotImplemented

    def __invert__(self):
        not_rq = ~ self._droprowsby_rq
        return type(self)(qualifier=not_rq)


# === Bound potential objects ===

class _DropRowsByColValColumnBoundPotential(object):

    def __init__(self, column_label: object) -> None:
        self.column_label = column_label

    # --- binary operators ---

    def __gt__(self, other):
        qualifier = rq.ColValGt(label=self.column_label, value=other)
        return DropRowsByQualifier(qualifier)

    def __ge__(self, other):
        qualifier = rq.ColValGe(label=self.column_label, value=other)
        return DropRowsByQualifier(qualifier)

    def __lt__(self, other):
        qualifier = rq.ColValLt(label=self.column_label, value=other)
        return DropRowsByQualifier(qualifier)

    def __le__(self, other):
        qualifier = rq.ColValLe(label=self.column_label, value=other)
        return DropRowsByQualifier(qualifier)

    def __eq__(self, other):
        qualifier = rq.ColValEq(label=self.column_label, value=other)
        return DropRowsByQualifier(qualifier)

    def __ne__(self, other):
        qualifier = rq.ColValNe(label=self.column_label, value=other)
        return DropRowsByQualifier(qualifier)

    # --- series methods ---

    def isin(
        self,
        value_list: Union[List[object], Set[object]],
    ) -> DropRowsByQualifier:
        q = rq.ColValIsIn(
            label=self.column_label,
            value_list=value_list,
        )
        return DropRowsByQualifier(q)

    def isna(self) -> DropRowsByQualifier:
        q = rq.ColValIsNa(self.column_label)
        return DropRowsByQualifier(q)

    def notna(self) -> DropRowsByQualifier:
        q = rq.ColValNotNa(self.column_label)
        return DropRowsByQualifier(q)


class _KeepRowsByColValColumnBoundPotential(object):

    def __init__(self, column_label: object) -> None:
        self.column_label = column_label

    # --- binary operators ---

    def __gt__(self, other):
        qualifier = rq.ColValGt(label=self.column_label, value=other)
        return KeepRowsByQualifier(qualifier)

    def __ge__(self, other):
        qualifier = rq.ColValGe(label=self.column_label, value=other)
        return KeepRowsByQualifier(qualifier)

    def __lt__(self, other):
        qualifier = rq.ColValLt(label=self.column_label, value=other)
        return KeepRowsByQualifier(qualifier)

    def __le__(self, other):
        qualifier = rq.ColValLe(label=self.column_label, value=other)
        return KeepRowsByQualifier(qualifier)

    def __eq__(self, other):
        qualifier = rq.ColValEq(label=self.column_label, value=other)
        return KeepRowsByQualifier(qualifier)

    def __ne__(self, other):
        qualifier = rq.ColValNe(label=self.column_label, value=other)
        return KeepRowsByQualifier(qualifier)

    # --- series methods ---

    def isin(
        self,
        value_list: Union[List[object], Set[object]],
    ) -> KeepRowsByQualifier:
        q = rq.ColValIsIn(
            label=self.column_label,
            value_list=value_list,
        )
        return KeepRowsByQualifier(q)

    def isna(self) -> KeepRowsByQualifier:
        q = rq.ColValIsNa(self.column_label)
        return KeepRowsByQualifier(q)

    def notna(self) -> KeepRowsByQualifier:
        q = rq.ColValNotNa(self.column_label)
        return KeepRowsByQualifier(q)


# === fly handle classes ===

class _DropRowsByColValHandle(object):

    def __getitem__(self, index):
        if isinstance(index, slice):
            raise ValueError("pdpipe fly handles are not meant for slicing!")
        return _DropRowsByColValColumnBoundPotential(index)


class _KeepRowsByColValHandle(object):

    def __getitem__(self, index):
        if isinstance(index, slice):
            raise ValueError("pdpipe fly handles are not meant for slicing!")
        return _KeepRowsByColValColumnBoundPotential(index)


# === fly handles ===

drop_rows_where = _DropRowsByColValHandle()

drop_rows_where.__doc__ = """
Use `drop_rows_where` as a handle to the future dataframe,
using the `[]` indexing syntax to select a (single) column to apply the logic
by, and regular binary operators such as >, >=, ==, !=, etc. to express the
condition by which rows will be kept or dropped.

For example:

>>> import pandas as pd; import pdpipe as pdp;
>>> df = pd.DataFrame([[1,4],[4,5],[5,11]], [1,2,3], ['a','b'])
>>> df
   a   b
1  1   4
2  4   5
3  5  11
>>> pipeline = pdp.PdPipeline([
...     pdp.drop_rows_where['a'] > 4,
... ])
>>> pipeline(df)
   a  b
1  1  4
2  4  5

The resulting stages can be naturaly combined by logical binary operators:
& for AND, | for OR and ^ for XOR, and can also be inverted with the `~`
operator.

For example:

>>> import pandas as pd; import pdpipe as pdp;
>>> pipeline = pdp.PdPipeline([
...     ~ (pdp.drop_rows_where['a'] > 4),
... ])
>>> pipeline(df)
   a   b
3  5  11
>>> pipeline = pdp.PdPipeline([
...     (pdp.drop_rows_where['a'] > 3) & (pdp.drop_rows_where['b'] < 10),
... ])
>>> pipeline(df)
   a   b
1  1   4
3  5  11

"""

keep_rows_where = _KeepRowsByColValHandle()



"""Test the RegexReplace pipeline stage."""

import re

import pandas as pd
import pdpipe as pdp


DF = pd.DataFrame(
    data=[[4, "more than 12"], [5, "with 5 more"]],
    index=[1, 2],
    columns=["age", "text"],
)


def test_regex_replace():
    clean_num = pdp.RegexReplace('text', r'\b[0-9]+\b', "NUM")
    res_df = clean_num(DF)
    assert 'age' in res_df.columns
    assert 'text' in res_df.columns
    assert res_df.loc[1]['text'] == 'more than NUM'
    assert res_df.loc[2]['text'] == 'with NUM more'


def test_regex_replace_no_drop():
    clean_num = pdp.RegexReplace('text', r'\b[0-9]+\b', "NUM", drop=False)
    res_df = clean_num(DF)
    assert 'age' in res_df.columns
    assert 'text' in res_df.columns
    assert 'text_regex' in res_df.columns
    assert res_df.loc[1]['text'] == 'more than 12'
    assert res_df.loc[2]['text'] == 'with 5 more'
    assert res_df.loc[1]['text_regex'] == 'more than NUM'
    assert res_df.loc[2]['text_regex'] == 'with NUM more'


DF2 = pd.DataFrame(
    data=[[4, "first\nsecond"], [5, "1 \n 2 \n 3"]],
    index=[1, 2],
    columns=["age", "text"],
)


DF3 = pd.DataFrame(
    data=[["Mr. John", 18], ["MR. Bob", 25]],
    index=[1, 2],
    columns=["name", "age"],
)


def test_regex_replace_with_flags():
    tokenizer = pdp.RegexReplace('text', r'.+', "TOKEN")
    res_df = tokenizer(DF2)
    assert 'age' in res_df.columns
    assert 'text' in res_df.columns
    assert res_df.loc[1]['text'] == 'TOKEN\nTOKEN'
    assert res_df.loc[2]['text'] == 'TOKEN\nTOKEN\nTOKEN'

    # now with the DOTALL flag
    tokenizer = pdp.RegexReplace('text', r'.+', "TOKEN", flags=re.DOTALL)
    res_df = tokenizer(DF2)
    assert 'age' in res_df.columns
    assert 'text' in res_df.columns
    assert res_df.loc[1]['text'] == 'TOKEN'
    assert res_df.loc[2]['text'] == 'TOKEN'

    # documentation example
    tokenizer = pdp.RegexReplace('name', r'^mr.*', "x", flags=re.IGNORECASE)
    res_df = tokenizer(DF3)
    assert 'name' in res_df.columns
    assert 'age' in res_df.columns
    assert res_df.loc[1]['name'] == 'x'
    assert res_df.loc[2]['name'] == 'x'
